# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_thread_protos/thread.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_tokenizer_proto import options_pb2 as pw__tokenizer__proto_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dpw_thread_protos/thread.proto\x12\x0fpw.thread.proto\x1a pw_tokenizer_proto/options.proto\"{\n\x0bThreadState\"l\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x15\n\x11INTERRUPT_HANDLER\x10\x01\x12\x0b\n\x07RUNNING\x10\x02\x12\t\n\x05READY\x10\x03\x12\r\n\tSUSPENDED\x10\x04\x12\x0b\n\x07\x42LOCKED\x10\x05\x12\x0c\n\x08INACTIVE\x10\x06\"\xc4\x03\n\x06Thread\x12\x12\n\x04name\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x12\n\x06\x61\x63tive\x18\x02 \x01(\x08\x42\x02\x18\x01\x12\x30\n\x05state\x18\x03 \x01(\x0e\x32!.pw.thread.proto.ThreadState.Enum\x12\x15\n\rraw_backtrace\x18\x04 \x03(\x04\x12\x1c\n\x14symbolized_backtrace\x18\x05 \x03(\t\x12\x11\n\traw_stack\x18\x06 \x01(\x0c\x12 \n\x13stack_start_pointer\x18\x07 \x01(\x04H\x00\x88\x01\x01\x12\x1e\n\x11stack_end_pointer\x18\x08 \x01(\x04H\x01\x88\x01\x01\x12\x1a\n\rstack_pointer\x18\t \x01(\x04H\x02\x88\x01\x01\x12!\n\x14\x63pu_usage_hundredths\x18\n \x01(\rH\x03\x88\x01\x01\x12#\n\x16stack_pointer_est_peak\x18\x0b \x01(\x04H\x04\x88\x01\x01\x42\x16\n\x14_stack_start_pointerB\x14\n\x12_stack_end_pointerB\x10\n\x0e_stack_pointerB\x17\n\x15_cpu_usage_hundredthsB\x19\n\x17_stack_pointer_est_peak\">\n\x12SnapshotThreadInfo\x12(\n\x07threads\x18\x12 \x03(\x0b\x32\x17.pw.thread.proto.ThreadB\x19\n\x0fpw.thread.protoB\x06Threadb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_thread_protos.thread_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\017pw.thread.protoB\006Thread'
  _globals['_THREAD'].fields_by_name['name']._options = None
  _globals['_THREAD'].fields_by_name['name']._serialized_options = b'\200\257&\000'
  _globals['_THREAD'].fields_by_name['active']._options = None
  _globals['_THREAD'].fields_by_name['active']._serialized_options = b'\030\001'
  _globals['_THREADSTATE']._serialized_start=84
  _globals['_THREADSTATE']._serialized_end=207
  _globals['_THREADSTATE_ENUM']._serialized_start=99
  _globals['_THREADSTATE_ENUM']._serialized_end=207
  _globals['_THREAD']._serialized_start=210
  _globals['_THREAD']._serialized_end=662
  _globals['_SNAPSHOTTHREADINFO']._serialized_start=664
  _globals['_SNAPSHOTTHREADINFO']._serialized_end=726
# @@protoc_insertion_point(module_scope)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for using ``pw_rpc.Client``."""

import logging
from typing import Any, Iterable

from pw_protobuf_compiler import python_protos
import pw_rpc
from pw_rpc import callback_client
from pw_stream import stream_readers

_LOG = logging.getLogger('pw_rpc.client_utils')
_VERBOSE = logging.DEBUG - 1


PathsModulesOrProtoLibrary = (
    Iterable[python_protos.PathOrModule] | python_protos.Library
)


class RpcClient:
    """An RPC client with configurable incoming data processing."""

    def __init__(
        self,
        reader_and_executor: stream_readers.DataReaderAndExecutor,
        paths_or_modules: PathsModulesOrProtoLibrary,
        channels: Iterable[pw_rpc.Channel],
        client_impl: pw_rpc.client.ClientImpl | None = None,
    ):
        """Creates an RPC client.

        Args:
          reader_and_executor: ``DataReaderAndExecutor`` instance.
          paths_or_modules: paths to .proto files or proto modules.
          channels: RPC channels to use for output.
          client_impl: The RPC client implementation. Defaults to the callback
            client implementation if not provided.
        """
        if isinstance(paths_or_modules, python_protos.Library):
            self.protos = paths_or_modules
        else:
            self.protos = python_protos.Library.from_paths(paths_or_modules)

        if client_impl is None:
            client_impl = callback_client.Impl()

        self.client = pw_rpc.Client.from_modules(
            client_impl, channels, self.protos.modules()
        )

        # Start background thread that reads and processes RPC packets.
        self._reader_and_executor = reader_and_executor
        self._reader_and_executor.start()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def close(self) -> None:
        self._reader_and_executor.stop()

    def rpcs(self, channel_id: int | None = None) -> Any:
        """Returns object for accessing services on the specified channel.

        This skips some intermediate layers to make it simpler to invoke RPCs
        from an ``HdlcRpcClient``. If only one channel is in use, the channel ID
        is not necessary.
        """
        if channel_id is None:
            return next(iter(self.client.channels())).rpcs

        return self.client.channel(channel_id).rpcs

    def handle_rpc_packet(self, packet: bytes) -> None:
        if not self.client.process_packet(packet):
            _LOG.error('Packet not handled by RPC client: %s', packet)


class NoEncodingSingleChannelRpcClient(RpcClient):
    """An RPC client without any frame encoding with a single channel output.

    The caveat is that the provided read function must read entire frames.
    """

    def __init__(
        self,
        reader: stream_readers.CancellableReader,
        paths_or_modules: PathsModulesOrProtoLibrary,
        channel: pw_rpc.Channel,
        client_impl: pw_rpc.client.ClientImpl | None = None,
    ):
        """Creates an RPC client over a single channel with no frame encoding.

        Args:
          reader: Readable object used to receive RPC packets.
          paths_or_modules: paths to .proto files or proto modules.
          channel: RPC channel to use for output.
          client_impl: The RPC Client implementation. Defaults to the callback
            client implementation if not provided.
        """

        def process_data(data: bytes):
            yield data

        def on_read_error(exc: Exception) -> None:
            _LOG.error('data reader encountered an error', exc_info=exc)

        reader_and_executor = stream_readers.DataReaderAndExecutor(
            reader, on_read_error, process_data, self.handle_rpc_packet
        )
        super().__init__(
            reader_and_executor, paths_or_modules, [channel], client_impl
        )



#!/usr/bin/env python
# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
""" Prompt toolkit application for pw watch. """

import asyncio
import functools
import logging
import os
import re
import time
from typing import Callable, Iterable, NoReturn

from prompt_toolkit.application import Application
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit.filters import Condition
from prompt_toolkit.history import (
    InMemoryHistory,
    History,
    ThreadedHistory,
)
from prompt_toolkit.key_binding import (
    KeyBindings,
    KeyBindingsBase,
    merge_key_bindings,
)
from prompt_toolkit.layout import (
    DynamicContainer,
    Float,
    FloatContainer,
    FormattedTextControl,
    HSplit,
    Layout,
    Window,
)
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.styles import (
    ConditionalStyleTransformation,
    DynamicStyle,
    SwapLightAndDarkStyleTransformation,
    merge_style_transformations,
    merge_styles,
    style_from_pygments_cls,
)
from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.markup import MarkdownLexer  # type: ignore

from pw_config_loader import yaml_config_loader_mixin

from pw_console.console_app import get_default_colordepth, MIN_REDRAW_INTERVAL
from pw_console.get_pw_console_app import PW_CONSOLE_APP_CONTEXTVAR
from pw_console.help_window import HelpWindow
from pw_console.key_bindings import DEFAULT_KEY_BINDINGS
from pw_console.log_pane import LogPane
from pw_console.plugin_mixin import PluginMixin
import pw_console.python_logging
from pw_console.quit_dialog import QuitDialog
from pw_console.style import generate_styles, get_theme_colors
from pw_console.pigweed_code_style import PigweedCodeStyle
from pw_console.widgets import (
    FloatingWindowPane,
    ToolbarButton,
    WindowPaneToolbar,
    create_border,
    mouse_handlers,
    to_checkbox,
)
from pw_console.window_list import DisplayMode
from pw_console.window_manager import WindowManager

from pw_build.project_builder_prefs import ProjectBuilderPrefs
from pw_build.project_builder_context import get_project_builder_context


_LOG = logging.getLogger('pw_build.watch')

BUILDER_CONTEXT = get_project_builder_context()

_HELP_TEXT = """
Mouse Keys
==========

- Click on a line in the bottom progress bar to switch to that tab.
- Click on any tab, or button to activate.
- Scroll wheel in the the log windows moves back through the history.


Global Keys
===========

Quit with confirmation dialog. --------------------  Ctrl-D
Quit without confirmation. ------------------------  Ctrl-X Ctrl-C
Toggle user guide window. -------------------------  F1
Trigger a rebuild. --------------------------------  Enter


Window Management Keys
======================

Switch focus to the next window pane or tab. ------  Ctrl-Alt-N
Switch focus to the previous window pane or tab. --  Ctrl-Alt-P
Move window pane left. ----------------------------  Ctrl-Alt-Left
Move window pane right. ---------------------------  Ctrl-Alt-Right
Move window pane down. ----------------------------  Ctrl-Alt-Down
Move window pane up. ------------------------------  Ctrl-Alt-Up
Balance all window sizes. -------------------------  Ctrl-U


Bottom Toolbar Controls
=======================

Rebuild Enter --------------- Click or press Enter to trigger a rebuild.
[x] Auto Rebuild ------------ Click to globaly enable or disable automatic
                              rebuilding when files change.
Help F1 --------------------- Click or press F1 to open this help window.
Quit Ctrl-d ----------------- Click or press Ctrl-d to quit pw_watch.
Next Tab Ctrl-Alt-n --------- Switch to the next log tab.
Previous Tab Ctrl-Alt-p ----- Switch to the previous log tab.


Build Status Bar
================

The build status bar shows the current status of all build directories outlined
in a colored frame.

  ┏━━ BUILDING ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
  ┃ [✓] out_directory  Building  Last line of standard out.                ┃
  ┃ [✓] out_dir2       Waiting   Last line of standard out.                ┃
  ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛

Each checkbox on the far left controls whether that directory is built when
files change and manual builds are run.


Copying Text
============

- Click drag will select whole lines in the log windows.
- `Ctrl-c` will copy selected lines to your system clipboard.

If running over SSH you will need to use your terminal's built in text
selection.

Linux
-----

- Holding `Shift` and dragging the mouse in most terminals.

Mac
---

- Apple Terminal:

  Hold `Fn` and drag the mouse

- iTerm2:

  Hold `Cmd+Option` and drag the mouse

Windows
-------

- Git CMD (included in `Git for Windows)

  1. Click on the Git window icon in the upper left of the title bar
  2. Click `Edit` then `Mark`
  3. Drag the mouse to select text and press Enter to copy.

- Windows Terminal

  1. Hold `Shift` and drag the mouse to select text
  2. Press `Ctrl-Shift-C` to copy.

"""


class WatchAppPrefs(ProjectBuilderPrefs):
    """Add pw_console specific prefs standard ProjectBuilderPrefs."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.registered_commands = DEFAULT_KEY_BINDINGS
        self.registered_commands.update(self.user_key_bindings)

        new_config_settings = {
            'key_bindings': DEFAULT_KEY_BINDINGS,
            'show_python_logger': True,
        }
        self.default_config.update(new_config_settings)
        self._update_config(
            new_config_settings,
            yaml_config_loader_mixin.Stage.DEFAULT,
        )

    # Required pw_console preferences for key bindings and themes
    @property
    def user_key_bindings(self) -> dict[str, list[str]]:
        return self._config.get('key_bindings', {})

    @property
    def ui_theme(self) -> str:
        return self._config.get('ui_theme', '')

    @ui_theme.setter
    def ui_theme(self, new_ui_theme: str) -> None:
        self._config['ui_theme'] = new_ui_theme

    @property
    def theme_colors(self):
        return get_theme_colors(self.ui_theme)

    @property
    def swap_light_and_dark(self) -> bool:
        return self._config.get('swap_light_and_dark', False)

    def get_function_keys(self, name: str) -> list:
        """Return the keys for the named function."""
        try:
            return self.registered_commands[name]
        except KeyError as error:
            raise KeyError('Unbound key function: {}'.format(name)) from error

    def register_named_key_function(
        self, name: str, default_bindings: list[str]
    ) -> None:
        self.registered_commands[name] = default_bindings

    def register_keybinding(
        self, name: str, key_bindings: KeyBindings, **kwargs
    ) -> Callable:
        """Apply registered keys for the given named function."""

        def decorator(handler: Callable) -> Callable:
            "`handler` is a callable or Binding."
            for keys in self.get_function_keys(name):
                key_bindings.add(*keys.split(' '), **kwargs)(handler)
            return handler

        return decorator

    # Required pw_console preferences for using a log window pane.
    @property
    def spaces_between_columns(self) -> int:
        return 2

    @property
    def window_column_split_method(self) -> str:
        return 'vertical'

    @property
    def hide_date_from_log_time(self) -> bool:
        return True

    @property
    def column_order(self) -> list:
        return []

    def column_style(  # pylint: disable=no-self-use
        self,
        _column_name: str,
        _column_value: str,
        default='',
    ) -> str:
        return default

    @property
    def show_python_file(self) -> bool:
        return self._config.get('show_python_file', False)

    @property
    def show_source_file(self) -> bool:
        return self._config.get('show_source_file', False)

    @property
    def show_python_logger(self) -> bool:
        return self._config.get('show_python_logger', False)


class WatchWindowManager(WindowManager):
    def update_root_container_body(self):
        self.application.window_manager_container = self.create_root_container()


class WatchApp(PluginMixin):
    """Pigweed Watch main window application."""

    # pylint: disable=too-many-instance-attributes

    NINJA_FAILURE_TEXT = '\033[31mFAILED: '

    NINJA_BUILD_STEP = re.compile(
        r"^\[(?P<step>[0-9]+)/(?P<total_steps>[0-9]+)\] (?P<action>.*)$"
    )

    def __init__(
        self,
        event_handler,
        prefs: WatchAppPrefs,
    ):
        self.event_handler = event_handler

        self.color_depth = get_default_colordepth()

        # Necessary for some of pw_console's window manager features to work
        # such as mouse drag resizing.
        PW_CONSOLE_APP_CONTEXTVAR.set(self)  # type: ignore

        self.prefs = prefs

        self.quit_dialog = QuitDialog(self, self.exit)  # type: ignore

        self.search_history: History = ThreadedHistory(InMemoryHistory())

        self.window_manager = WatchWindowManager(self)

        self._build_error_count = 0
        self._errors_in_output = False

        self.log_ui_update_frequency = 0.1  # 10 FPS
        self._last_ui_update_time = time.time()

        self.recipe_name_to_log_pane: dict[str, LogPane] = {}
        self.recipe_index_to_log_pane: dict[int, LogPane] = {}

        debug_logging = (
            event_handler.project_builder.default_log_level == logging.DEBUG
        )
        level_name = 'DEBUG' if debug_logging else 'INFO'

        no_propagation_loggers = []

        if event_handler.separate_logfiles:
            pane_index = len(event_handler.project_builder.build_recipes) - 1
            for recipe in reversed(event_handler.project_builder.build_recipes):
                log_pane = self.add_build_log_pane(
                    recipe.display_name,
                    loggers=[recipe.log],
                    level_name=level_name,
                )
                if recipe.log.propagate is False:
                    no_propagation_loggers.append(recipe.log)

                self.recipe_name_to_log_pane[recipe.display_name] = log_pane
                self.recipe_index_to_log_pane[pane_index] = log_pane
                pane_index -= 1

        pw_console.python_logging.setup_python_logging(
            loggers_with_no_propagation=no_propagation_loggers
        )

        self.root_log_pane = self.add_build_log_pane(
            'Root Log',
            loggers=[
                logging.getLogger('pw_build'),
            ],
            level_name=level_name,
        )
        # Repeat the Attaching filesystem watcher message for the full screen
        # interface. The original log in watch.py will be hidden from view.
        _LOG.info('Attaching filesystem watcher...')

        self.window_manager.window_lists[0].display_mode = DisplayMode.TABBED

        self.window_manager_container = (
            self.window_manager.create_root_container()
        )

        self.status_bar_border_style = 'class:command-runner-border'

        self.status_bar_control = FormattedTextControl(self.get_status_bar_text)

        self.status_bar_container = create_border(
            HSplit(
                [
                    # Result Toolbar.
                    Window(
                        content=self.status_bar_control,
                        height=len(self.event_handler.project_builder),
                        wrap_lines=False,
                        style='class:pane_active',
                    ),
                ]
            ),
            content_height=len(self.event_handler.project_builder),
            title=BUILDER_CONTEXT.get_title_bar_text,
            border_style=(BUILDER_CONTEXT.get_title_style),
            base_style='class:pane_active',
            left_margin_columns=1,
            right_margin_columns=1,
        )

        self.floating_window_plugins: list[FloatingWindowPane] = []

        self.user_guide_window = HelpWindow(
            self,  # type: ignore
            title='Pigweed Watch',
            disable_ctrl_c=True,
        )
        self.user_guide_window.set_help_text(
            _HELP_TEXT, lexer=PygmentsLexer(MarkdownLexer)
        )

        self.help_toolbar = WindowPaneToolbar(
            title='Pigweed Watch',
            include_resize_handle=False,
            focus_action_callable=self.switch_to_root_log,
            click_to_focus_text='',
        )
        self.help_toolbar.add_button(
            ToolbarButton('Enter', 'Rebuild', self.run_build)
        )
        self.help_toolbar.add_button(
            ToolbarButton(
                description='Auto Rebuild',
                mouse_handler=self.toggle_restart_on_filechange,
                is_checkbox=True,
                checked=lambda: self.restart_on_changes,
            )
        )
        self.help_toolbar.add_button(
            ToolbarButton('F1', 'Help', self.user_guide_window.toggle_display)
        )
        self.help_toolbar.add_button(ToolbarButton('Ctrl-d', 'Quit', self.exit))
        self.help_toolbar.add_button(
            ToolbarButton(
                'Ctrl-Alt-n', 'Next Tab', self.window_manager.focus_next_pane
            )
        )
        self.help_toolbar.add_button(
            ToolbarButton(
                'Ctrl-Alt-p',
                'Previous Tab',
                self.window_manager.focus_previous_pane,
            )
        )

        self.root_container = FloatContainer(
            HSplit(
                [
                    # Window pane content:
                    DynamicContainer(lambda: self.window_manager_container),
                    self.status_bar_container,
                    self.help_toolbar,
                ]
            ),
            floats=[
                Float(
                    content=self.user_guide_window,
                    top=2,
                    left=4,
                    bottom=4,
                    width=self.user_guide_window.content_width,
                ),
                Float(
                    content=self.quit_dialog,
                    top=2,
                    left=2,
                ),
            ],
        )

        key_bindings = KeyBindings()

        @key_bindings.add('enter', filter=self.input_box_not_focused())
        def _run_build(_event):
            "Rebuild."
            self.run_build()

        register = self.prefs.register_keybinding

        @register('global.exit-no-confirmation', key_bindings)
        def _quit_no_confirm(_event):
            """Quit without confirmation."""
            _LOG.info('Got quit signal; exiting...')
            self.exit(0)

        @register('global.exit-with-confirmation', key_bindings)
        def _quit_with_confirm(_event):
            """Quit with confirmation dialog."""
            self.quit_dialog.open_dialog()

        @register(
            'global.open-user-guide',
            key_bindings,
            filter=Condition(lambda: not self.modal_window_is_open()),
        )
        def _show_help(_event):
            """Toggle user guide window."""
            self.user_guide_window.toggle_display()

        self.key_bindings = merge_key_bindings(
            [
                self.window_manager.key_bindings,
                key_bindings,
            ]
        )

        self.current_theme = generate_styles(self.prefs.ui_theme)

        self.style_transformation = merge_style_transformations(
            [
                ConditionalStyleTransformation(
                    SwapLightAndDarkStyleTransformation(),
                    filter=Condition(lambda: self.prefs.swap_light_and_dark),
                ),
            ]
        )

        self.code_theme = style_from_pygments_cls(PigweedCodeStyle)

        self.layout = Layout(
            self.root_container,
            focused_element=self.root_log_pane,
        )

        self.application: Application = Application(
            layout=self.layout,
            key_bindings=self.key_bindings,
            mouse_support=True,
            color_depth=self.color_depth,
            clipboard=PyperclipClipboard(),
            style=DynamicStyle(
                lambda: merge_styles(
                    [
                        self.current_theme,
                        self.code_theme,
                    ]
                )
            ),
            style_transformation=self.style_transformation,
            full_screen=True,
            min_redraw_interval=MIN_REDRAW_INTERVAL,
        )

        self.plugin_init(
            plugin_callback=self.check_build_status,
            plugin_callback_frequency=0.5,
            plugin_logger_name='pw_watch_stdout_checker',
        )

    def add_build_log_pane(
        self, title: str, loggers: list[logging.Logger], level_name: str
    ) -> LogPane:
        """Setup a new build log pane."""
        new_log_pane = LogPane(application=self, pane_title=title)
        for logger in loggers:
            new_log_pane.add_log_handler(logger, level_name=level_name)

        # Set python log format to just the message itself.
        new_log_pane.log_view.log_store.formatter = logging.Formatter(
            '%(message)s'
        )

        new_log_pane.table_view = False

        # Disable line wrapping for improved error visibility.
        if new_log_pane.wrap_lines:
            new_log_pane.toggle_wrap_lines()

        # Blank right side toolbar text
        new_log_pane._pane_subtitle = ' '  # pylint: disable=protected-access

        # Make tab and shift-tab search for next and previous error
        next_error_bindings = KeyBindings()

        @next_error_bindings.add('s-tab')
        def _previous_error(_event):
            self.jump_to_error(backwards=True)

        @next_error_bindings.add('tab')
        def _next_error(_event):
            self.jump_to_error()

        existing_log_bindings: (
            KeyBindingsBase | None
        ) = new_log_pane.log_content_control.key_bindings

        key_binding_list: list[KeyBindingsBase] = []
        if existing_log_bindings:
            key_binding_list.append(existing_log_bindings)
        key_binding_list.append(next_error_bindings)
        new_log_pane.log_content_control.key_bindings = merge_key_bindings(
            key_binding_list
        )

        # Only show a few buttons in the log pane toolbars.
        new_buttons = []
        for button in new_log_pane.bottom_toolbar.buttons:
            if button.description in [
                'Search',
                'Save',
                'Follow',
                'Wrap',
                'Clear',
            ]:
                new_buttons.append(button)
        new_log_pane.bottom_toolbar.buttons = new_buttons

        self.window_manager.add_pane(new_log_pane)
        return new_log_pane

    def logs_redraw(self):
        emit_time = time.time()
        # Has enough time passed since last UI redraw due to new logs?
        if emit_time > self._last_ui_update_time + self.log_ui_update_frequency:
            # Update last log time
            self._last_ui_update_time = emit_time

            # Trigger Prompt Toolkit UI redraw.
            self.redraw_ui()

    def jump_to_error(self, backwards: bool = False) -> None:
        if not self.root_log_pane.log_view.search_text:
            self.root_log_pane.log_view.set_search_regex(
                '^FAILED: ', False, None
            )
        if backwards:
            self.root_log_pane.log_view.search_backwards()
        else:
            self.root_log_pane.log_view.search_forwards()
        self.root_log_pane.log_view.log_screen.reset_logs(
            log_index=self.root_log_pane.log_view.log_index
        )

        self.root_log_pane.log_view.move_selected_line_to_top()

    def refresh_layout(self) -> None:
        self.window_manager.update_root_container_body()

    def update_menu_items(self):
        """Required by the Window Manager Class."""

    def redraw_ui(self):
        """Redraw the prompt_toolkit UI."""
        if hasattr(self, 'application'):
            self.application.invalidate()

    def focus_on_container(self, pane):
        """Set application focus to a specific container."""
        # Try to focus on the given pane
        try:
            self.application.layout.focus(pane)
        except ValueError:
            # If the container can't be focused, focus on the first visible
            # window pane.
            self.window_manager.focus_first_visible_pane()

    def focused_window(self):
        """Return the currently focused window."""
        return self.application.layout.current_window

    def focus_main_menu(self):
        """Focus on the main menu.

        Currently pw_watch has no main menu so focus on the first visible pane
        instead."""
        self.window_manager.focus_first_visible_pane()

    def switch_to_root_log(self) -> None:
        (
            window_list,
            pane_index,
        ) = self.window_manager.find_window_list_and_pane_index(
            self.root_log_pane
        )
        window_list.switch_to_tab(pane_index)

    def switch_to_build_log(self, log_index: int) -> None:
        pane = self.recipe_index_to_log_pane.get(log_index, None)
        if not pane:
            return

        (
            window_list,
            pane_index,
        ) = self.window_manager.find_window_list_and_pane_index(pane)
        window_list.switch_to_tab(pane_index)

    def command_runner_is_open(self) -> bool:
        # pylint: disable=no-self-use
        return False

    def all_log_panes(self) -> Iterable[LogPane]:
        for pane in self.window_manager.active_panes():
            if isinstance(pane, LogPane):
                yield pane

    def clear_log_panes(self) -> None:
        """Erase all log pane content and turn on follow.

        This is called whenever rebuilds occur. Either a manual build from
        self.run_build or on file changes called from
        pw_watch._handle_matched_event."""
        for pane in self.all_log_panes():
            pane.log_view.clear_visual_selection()
            pane.log_view.clear_filters()
            pane.log_view.log_store.clear_logs()
            pane.log_view.view_mode_changed()
            # Re-enable follow if needed
            if not pane.log_view.follow:
                pane.log_view.toggle_follow()

    def run_build(self) -> None:
        """Manually trigger a rebuild from the UI."""
        self.clear_log_panes()
        self.event_handler.rebuild()

    @property
    def restart_on_changes(self) -> bool:
        return self.event_handler.restart_on_changes

    def toggle_restart_on_filechange(self) -> None:
        self.event_handler.restart_on_changes = (
            not self.event_handler.restart_on_changes
        )

    def get_status_bar_text(self) -> StyleAndTextTuples:
        """Return formatted text for build status bar."""
        formatted_text: StyleAndTextTuples = []

        separator = ('', ' ')
        name_width = self.event_handler.project_builder.max_name_width

        # pylint: disable=protected-access
        (
            _window_list,
            pane,
        ) = self.window_manager._get_active_window_list_and_pane()
        # pylint: enable=protected-access
        restarting = BUILDER_CONTEXT.restart_flag

        for i, cfg in enumerate(self.event_handler.project_builder):
            # The build directory
            name_style = ''
            if not pane:
                formatted_text.append(('', '\n'))
                continue

            # Dim the build name if disabled
            if not cfg.enabled:
                name_style = 'class:theme-fg-inactive'

            # If this build tab is selected, highlight with cyan.
            if pane.pane_title() == cfg.display_name:
                name_style = 'class:theme-fg-cyan'

            formatted_text.append(
                to_checkbox(
                    cfg.enabled,
                    functools.partial(
                        mouse_handlers.on_click,
                        cfg.toggle_enabled,
                    ),
                    end=' ',
                    unchecked_style='class:checkbox',
                    checked_style='class:checkbox-checked',
                )
            )
            formatted_text.append(
                (
                    name_style,
                    f'{cfg.display_name}'.ljust(name_width),
                    functools.partial(
                        mouse_handlers.on_click,
                        functools.partial(self.switch_to_build_log, i),
                    ),
                )
            )
            formatted_text.append(separator)
            # Status
            formatted_text.append(cfg.status.status_slug(restarting=restarting))
            formatted_text.append(separator)
            # Current stdout line
            formatted_text.extend(cfg.status.current_step_formatted())
            formatted_text.append(('', '\n'))

        if not formatted_text:
            formatted_text = [('', 'Loading...')]

        self.set_tab_bar_colors()

        return formatted_text

    def set_tab_bar_colors(self) -> None:
        restarting = BUILDER_CONTEXT.restart_flag

        for cfg in BUILDER_CONTEXT.recipes:
            pane = self.recipe_name_to_log_pane.get(cfg.display_name, None)
            if not pane:
                continue

            pane.extra_tab_style = None
            if not restarting and cfg.status.failed():
                pane.extra_tab_style = 'class:theme-fg-red'

    def exit(
        self,
        exit_code: int = 1,
        log_after_shutdown: Callable[[], None] | None = None,
    ) -> None:
        _LOG.info('Exiting...')
        BUILDER_CONTEXT.ctrl_c_pressed = True

        # Shut everything down after the prompt_toolkit app exits.
        def _really_exit(future: asyncio.Future) -> NoReturn:
            BUILDER_CONTEXT.restore_logging_and_shutdown(log_after_shutdown)
            os._exit(future.result())  # pylint: disable=protected-access

        if self.application.future:
            self.application.future.add_done_callback(_really_exit)
        self.application.exit(result=exit_code)

    def check_build_status(self) -> bool:
        if not self.event_handler.current_stdout:
            return False

        if self._errors_in_output:
            return True

        if self.event_handler.current_build_errors > self._build_error_count:
            self._errors_in_output = True
            self.jump_to_error()

        return True

    def run(self):
        self.plugin_start()
        # Run the prompt_toolkit application
        self.application.run(set_exception_handler=True)

    def input_box_not_focused(self) -> Condition:
        """Condition checking the focused control is not a text input field."""

        @Condition
        def _test() -> bool:
            """Check if the currently focused control is an input buffer.

            Returns:
                bool: True if the currently focused control is not a text input
                    box. For example if the user presses enter when typing in
                    the search box, return False.
            """
            return not isinstance(
                self.application.layout.current_control, BufferControl
            )

        return _test

    def modal_window_is_open(self):
        """Return true if any modal window or dialog is open."""
        floating_window_is_open = (
            self.user_guide_window.show_window or self.quit_dialog.show_dialog
        )

        floating_plugin_is_open = any(
            plugin.show_pane for plugin in self.floating_window_plugins
        )

        return floating_window_is_open or floating_plugin_is_open



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Default pw build script for upstream Pigweed.

Usage:

   pw build --list
   pw build --recipe default_gn
   pw build --recipe default_* --watch
   pw build --step gn_combined_build_check --step gn_python_*
   pw build -C out/gn --watch
"""

import argparse
import logging
from pathlib import Path
import shutil
import sys

import pw_cli.env
import pw_presubmit.pigweed_presubmit
from pw_presubmit.build import gn_args

from pw_build.build_recipe import (
    BuildCommand,
    BuildRecipe,
    should_gn_gen,
    should_regenerate_cmake,
)
from pw_build.project_builder_presubmit_runner import (
    get_parser,
    main,
)


_LOG = logging.getLogger('pw_build')

_PW_ENV = pw_cli.env.pigweed_environment()
_REPO_ROOT = _PW_ENV.PW_PROJECT_ROOT
_PACKAGE_ROOT = _PW_ENV.PW_PACKAGE_ROOT


def gn_recipe() -> BuildRecipe:
    """Return the default_gn recipe."""
    default_gn_gen_command = [
        'gn',
        'gen',
        # NOTE: Not an f-string. BuildRecipe will replace with the out dir.
        '{build_dir}',
        '--export-compile-commands',
    ]

    if shutil.which('ccache'):
        default_gn_gen_command.append(gn_args(pw_command_launcher='ccache'))

    return BuildRecipe(
        build_dir=Path('out/gn'),
        title='default_gn',
        steps=[
            BuildCommand(
                run_if=should_gn_gen,
                command=default_gn_gen_command,
            ),
            BuildCommand(
                build_system_command='ninja',
                targets=['default'],
            ),
        ],
    )


def bazel_recipe() -> BuildRecipe:
    """Return the default_bazel recipe."""
    default_bazel_targets = ['//...:all']

    return BuildRecipe(
        build_dir=Path('out/bazel'),
        title='default_bazel',
        steps=[
            BuildCommand(
                build_system_command='bazel',
                build_system_extra_args=[
                    'build',
                    '--verbose_failures',
                    '--worker_verbose',
                ],
                targets=default_bazel_targets,
            ),
            BuildCommand(
                build_system_command='bazel',
                build_system_extra_args=[
                    'test',
                    '--test_output=errors',
                ],
                targets=default_bazel_targets,
            ),
        ],
    )


def cmake_recipe() -> BuildRecipe:
    """Construct the default_cmake recipe."""
    toolchain_path = (
        _REPO_ROOT / 'pw_toolchain' / 'host_clang' / 'toolchain.cmake'
    )

    cmake_generate_command = [
        'cmake',
        '--fresh',
        '--debug-output',
        '-DCMAKE_MESSAGE_LOG_LEVEL=WARNING',
        '-S',
        str(_REPO_ROOT),
        '-B',
        # NOTE: Not an f-string. BuildRecipe will replace with the out dir.
        '{build_dir}',
        '-G',
        'Ninja',
        f'-DCMAKE_TOOLCHAIN_FILE={toolchain_path}',
        '-DCMAKE_EXPORT_COMPILE_COMMANDS=1',
        f'-Ddir_pw_third_party_nanopb={_PACKAGE_ROOT / "nanopb"}',
        '-Dpw_third_party_nanopb_ADD_SUBDIRECTORY=ON',
        f'-Ddir_pw_third_party_emboss={_PACKAGE_ROOT / "emboss"}',
        f'-Ddir_pw_third_party_boringssl={_PACKAGE_ROOT / "boringssl"}',
    ]

    if shutil.which('ccache'):
        cmake_generate_command.append("-DCMAKE_C_COMPILER_LAUNCHER=ccache")
        cmake_generate_command.append("-DCMAKE_CXX_COMPILER_LAUNCHER=ccache")

    pw_package_install_steps = [
        BuildCommand(
            command=['pw', '--no-banner', 'package', 'install', package],
        )
        for package in ['emboss', 'nanopb', 'boringssl']
    ]

    return BuildRecipe(
        build_dir=Path('out/cmake'),
        title='default_cmake',
        steps=pw_package_install_steps
        + [
            BuildCommand(
                run_if=should_regenerate_cmake(cmake_generate_command),
                command=cmake_generate_command,
            ),
            BuildCommand(
                build_system_command='ninja',
                targets=pw_presubmit.pigweed_presubmit.CMAKE_TARGETS,
            ),
        ],
    )


def pigweed_upstream_main() -> int:
    """Entry point for Pigweed upstream ``pw build`` command.

    Defines one or more BuildRecipes and passes that along with all of Pigweed
    upstream presubmit programs to the project_builder_presubmit_runner.main to
    start a pw build invocation.

    Returns:
      An int representing the success or failure status of the build; 0 if
      successful, 1 if failed.

    Command line usage examples:

    .. code-block:: bash

       pw build --list
       pw build --recipe default_gn
       pw build --recipe default_* --watch
       pw build --step gn_combined_build_check --step gn_python_*
       pw build -C out/gn --watch
    """

    return main(
        presubmit_programs=pw_presubmit.pigweed_presubmit.PROGRAMS,
        build_recipes=[
            gn_recipe(),
            bazel_recipe(),
            cmake_recipe(),
        ],
        default_root_logfile=Path('out/build.txt'),
    )


def _build_argument_parser() -> argparse.ArgumentParser:
    return get_parser(
        pw_presubmit.pigweed_presubmit.PROGRAMS,
        [
            gn_recipe(),
            bazel_recipe(),
            cmake_recipe(),
        ],
    )


if __name__ == '__main__':
    sys.exit(pigweed_upstream_main())



#!/usr/bin/env python3

# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Applies clang-tidy suggested fixes.

This uses the clang-apply-replacements tool to apply code changes described in
a YAML file generated by clang-tidy.
"""

import argparse
import logging
import re
from pathlib import Path
from pw_cli.tool_runner import BasicSubprocessRunner

_LOG: logging.Logger = logging.getLogger(__name__)
_REMOVE_CHANGE_FILES_FLAG = "--remove-change-desc-files"
_IGNORE_INSERT_CONFLICT_FLAG = "--ignore-insert-conflict"
_CLANG_TIDY_SUGGESTED_FIX_FILE_REGEX = "*.o.yaml"
_PIGWEED_INCLUDE_FIX_REGEX = re.compile(r'ReplacementText: "#include <(pw_.*)>')
_PIGWEED_INCLUDE_REPLACEMENT_REGEX = r'ReplacementText: "#include \"\1\"'


def apply_replacements(
    root: Path, remove_change_desc_files: bool, raise_insert_conflict: bool
) -> int:
    """Runs the clang-apply-replacements tool to apply clang-tidy fixes.

    Args:
        root: The directory under which clang-apply-replacements will search
            for YAML fix files generated by clang-tidy.
        remove_change_desc_files: Whether or not to remove the change
            description files regardless of successful.
        raise_insert_conflict: Whether or not to ignore insert conflicts.

    Returns:
        The return code of the clang-apply-replacements tool.
    """
    # Search under root for YAML files generated by clang-tidy that contain
    # suggested fixes
    for filepath in root.rglob(_CLANG_TIDY_SUGGESTED_FIX_FILE_REGEX):
        clang_tidy_fixes = filepath.read_text()

        # Change suggested pigweed includes from <pw_.*> to "pw_.*"
        if clang_tidy_fixes:
            new_clang_tidy_fixes = _PIGWEED_INCLUDE_FIX_REGEX.sub(
                _PIGWEED_INCLUDE_REPLACEMENT_REGEX, clang_tidy_fixes
            )
            filepath.write_text(new_clang_tidy_fixes)

    # Add flags for the clang-apply-replacements tool
    flags = []
    if remove_change_desc_files:
        flags.append(_REMOVE_CHANGE_FILES_FLAG)
    if raise_insert_conflict:
        flags.append(_IGNORE_INSERT_CONFLICT_FLAG)

    # Use the clang-apply-replacements tool via a subprocess
    _LOG.info('Applying clang-tidy fixes')
    run_tool = BasicSubprocessRunner()
    process_result = run_tool(
        'clang-apply-replacements',
        [str(root)] + flags,
        check=True,
    )
    return process_result.returncode


def arguments() -> argparse.ArgumentParser:
    """Creates an argument parser for clang-apply-replacements tool."""

    parser = argparse.ArgumentParser(description=__doc__)

    def existing_path(arg: str) -> Path:
        path = Path(arg)
        if not path.is_dir():
            raise argparse.ArgumentTypeError(
                f'{arg} is not a path to a directory'
            )

        return path

    parser.add_argument(
        'root',
        type=existing_path,
        help=(
            'Root directory that clang-apply-replacements will recursively '
            'search under for the YAML fix files generated by clang-tidy.'
        ),
    )

    parser.add_argument(
        '--remove-change-desc-files',
        action=argparse.BooleanOptionalAction,
        default=True,
        help=(
            'Remove the change description files regardless of successful '
            'merging/replacing.'
        ),
    )

    parser.add_argument(
        '--raise-insert-conflict',
        action=argparse.BooleanOptionalAction,
        default=True,
        help='Do not ignore insert conflicts.',
    )

    return parser


def main() -> int:
    """Check and fix formatting for source files."""
    return apply_replacements(**vars(arguments().parse_args()))



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""NPM pw_env_setup action plugin.

This action triggers `npm install` after CIPD setup.
"""
import os
import shutil
import subprocess


def run_action(env=None):
    """Project action to run 'npm install'."""
    if not env:
        raise ValueError(f"Missing 'env', got %{env}")

    with env():
        npm = shutil.which('npm.cmd' if os.name == 'nt' else 'npm')

        repo_root = os.environ.get('PW_PROJECT_ROOT') or os.environ.get(
            'PW_ROOT'
        )

        # TODO: b/323378974 - Better integrate NPM actions with pw_env_setup so
        # we don't have to manually set `npm_config_cache` every time we run
        # npm.
        # Force npm cache to live inside the environment directory.
        npm_env = os.environ.copy()
        npm_env['npm_config_cache'] = os.path.join(
            npm_env['_PW_ACTUAL_ENVIRONMENT_ROOT'], 'npm-cache'
        )

        subprocess.run(
            [
                npm,
                "install",
                "--quiet",
                "--no-progress",
                "--loglevel=error",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.DEVNULL,
            cwd=repo_root,
            env=npm_env,
        )

        subprocess.run(
            [
                npm,
                "run",
                "log-viewer-setup",
                "--quiet",
                "--no-progress",
                "--loglevel=error",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.DEVNULL,
            cwd=repo_root,
            env=npm_env,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_snapshot_protos/snapshot.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_chrono_protos import chrono_pb2 as pw__chrono__protos_dot_chrono__pb2
from pw_cpu_exception_cortex_m_protos import cpu_state_pb2 as pw__cpu__exception__cortex__m__protos_dot_cpu__state__pb2
from pw_log.proto import log_pb2 as pw__log_dot_proto_dot_log__pb2
from pw_thread_protos import thread_pb2 as pw__thread__protos_dot_thread__pb2
from pw_snapshot_metadata_proto import snapshot_metadata_pb2 as pw__snapshot__metadata__proto_dot_snapshot__metadata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!pw_snapshot_protos/snapshot.proto\x12\x0bpw.snapshot\x1a\x1dpw_chrono_protos/chrono.proto\x1a\x30pw_cpu_exception_cortex_m_protos/cpu_state.proto\x1a\x16pw_log/proto/log.proto\x1a\x1dpw_thread_protos/thread.proto\x1a\x32pw_snapshot_metadata_proto/snapshot_metadata.proto\"\xac\x03\n\x08Snapshot\x12\x1e\n\x04logs\x18\x01 \x03(\x0b\x32\x10.pw.log.LogEntry\x12\'\n\x08metadata\x18\x10 \x01(\x0b\x32\x15.pw.snapshot.Metadata\x12-\n\x04tags\x18\x11 \x03(\x0b\x32\x1f.pw.snapshot.Snapshot.TagsEntry\x12(\n\x07threads\x18\x12 \x03(\x0b\x32\x17.pw.thread.proto.Thread\x12\x30\n\x11related_snapshots\x18\x13 \x03(\x0b\x32\x15.pw.snapshot.Snapshot\x12\x43\n\x10\x61rmv7m_cpu_state\x18\x14 \x01(\x0b\x32).pw.cpu_exception.cortex_m.ArmV7mCpuState\x12\x12\n\ntrace_data\x18\x15 \x01(\x0c\x12(\n\ntimestamps\x18\x16 \x03(\x0b\x32\x14.pw.chrono.TimePoint\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01J\x04\x08\x02\x10\x08J\x04\x08\x08\x10\x10J\x05\x08\x17\x10\x88\x08J\t\x08\x88\x08\x10\x80\x80\x80\x80\x02\x42\x1d\n\x11pw.snapshot.protoB\x08Snapshotb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_snapshot_protos.snapshot_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\021pw.snapshot.protoB\010Snapshot'
  _globals['_SNAPSHOT_TAGSENTRY']._options = None
  _globals['_SNAPSHOT_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_SNAPSHOT']._serialized_start=239
  _globals['_SNAPSHOT']._serialized_end=667
  _globals['_SNAPSHOT_TAGSENTRY']._serialized_start=594
  _globals['_SNAPSHOT_TAGSENTRY']._serialized_end=637
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Wrapper script to run arm-none-eabi-gdb with Python 3.8."""

import os
from pathlib import Path
import shutil
import signal
import sys
import subprocess


def main() -> int:
    """arm-gdb wrapper that sets up the Python environment for gdb"""

    # Find 'arm-none-eabi-gdb' as long as it isn't in the current Python
    # virtualenv entry point. In other words: not this script.
    exclude_paths = sys.path
    venv = os.environ.get('VIRTUAL_ENV')
    if venv:
        venv_path = Path(venv).resolve()
        exclude_paths.append(os.path.join(venv_path, 'Scripts'))
    arm_gdb_binary = shutil.which(
        'arm-none-eabi-gdb',
        path=os.pathsep.join(
            [
                path_entry
                for path_entry in os.environ.get('PATH', '').split(os.pathsep)
                if str(Path(path_entry).resolve()) not in exclude_paths
            ]
        ),
    )
    assert arm_gdb_binary

    arm_gdb_path = Path(arm_gdb_binary)
    arm_install_prefix = arm_gdb_path.parent.parent
    python_home = arm_install_prefix / 'python/bin/python3'
    python_path = arm_install_prefix / 'python/lib/python3.8'
    assert arm_gdb_path.is_file()

    env = os.environ.copy()
    # Only set Python if it's in the expected location.
    if python_home.is_file() and python_path.is_dir():
        env['PYTHONHOME'] = str(python_home)
        env['PYTHONPATH'] = str(python_path)

    # Ignore Ctrl-C to allow gdb to handle normally
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    return subprocess.run(
        [str(arm_gdb_path)] + sys.argv[1:], env=env, check=False
    ).returncode


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Interactive input prompts with up/down navigation and validation."""

import sys

from prompt_toolkit import PromptSession
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.validation import Validator


def interactive_index_select(
    selection_lines: list[str],
    prompt_text: str = (
        '\nEnter an item index or press up/down (Ctrl-C to cancel)\n> '
    ),
) -> tuple[int, str]:
    """Display an interactive prompt to the user with up down navigation.

    Presents an interactive prompt to the user on the terminal where
    the index of and item can be typed in or items selected with the
    up & down keys. If the user enters Ctrl-c or Ctrl-d the process
    will exit with sys.exit(1).

    Args:
      selection_lines: List of strings to present to the user for
          selection. Indexes will be prepended to each line starting
          from 1 before displaying to the user.
      prompt_text: String to display before the prompt cursor.

    Returns: a tuple containing the index of the item selected and the
      selected item text.
    """

    lines_with_indexes = list(
        f'{i+1} - {line}' for i, line in enumerate(selection_lines)
    )

    # Add the items to a prompt_toolkit history so they can be
    # selected with up and down.
    history = InMemoryHistory()
    # Add in reverse order so pressing up selects the first item
    # followed by the second. Otherwise pressing up selects the last
    # item in the printed list.
    for line in reversed(lines_with_indexes):
        history.append_string(line)

    for line in lines_with_indexes:
        print(' ', line)

    def input_index(text: str) -> int:
        """Convert input to a single index.

        Args:
            text: str beginning a single integer followed by whitespace.

        Returns:
            The single integer as an int minus 1 for use as a list index.
        """
        parts = text.split()
        index = int(parts[0].strip())
        return index - 1

    def is_valid_item(text: str) -> bool:
        """Returns true if the input line is a valid entry."""
        return (
            bool(text)
            and any(line.startswith(text) for line in lines_with_indexes)
            and input_index(text) < len(lines_with_indexes)
        )

    validator = Validator.from_callable(
        is_valid_item,
        error_message="Not a valid item.",
        move_cursor_to_end=True,
    )

    session: PromptSession = PromptSession(
        history=history,
        enable_history_search=True,
        validator=validator,
        validate_while_typing=False,
    )

    try:
        selected_text = session.prompt(prompt_text)
    except (EOFError, KeyboardInterrupt):
        # Cancel with Ctrl-c or Ctrl-d
        sys.exit(1)

    selected_index = input_index(selected_text)
    original_text = selection_lines[selected_index]

    return (selected_index, original_text)



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
bloat is a script which generates a size report card for binary files.
"""

import argparse
import json
import logging
import os
from pathlib import Path
import subprocess
import sys
import tempfile
from typing import Iterable

import pw_cli.log

from pw_bloat.bloaty_config import generate_bloaty_config
from pw_bloat.label import DataSourceMap
from pw_bloat.label_output import (
    BloatTableOutput,
    LineCharset,
    RstOutput,
    AsciiCharset,
)

_LOG = logging.getLogger(__name__)

MAX_COL_WIDTH = 50
BINARY_SIZES_EXTENSION = '.binary_sizes.json'


def parse_args() -> argparse.Namespace:
    """Parses the script's arguments."""

    parser = argparse.ArgumentParser('Generate a size report card for binaries')
    parser.add_argument(
        '--gn-arg-path',
        type=str,
        required=True,
        help='File path to json of binaries',
    )
    parser.add_argument(
        '--single-report',
        action="store_true",
        help='Determine if calling single size report',
    )
    parser.add_argument(
        '--json-key-prefix',
        type=str,
        help='Prefix for json keys in size report, default = target name',
        default=None,
    )
    parser.add_argument(
        '--full-json-summary',
        action="store_true",
        help='Include all levels of data sources in json binary report',
    )
    parser.add_argument(
        '--ignore-unused-labels',
        action="store_true",
        help='Do not include labels with size equal to zero in report',
    )

    return parser.parse_args()


def run_bloaty(
    filename: str,
    config: str,
    base_file: str | None = None,
    data_sources: Iterable[str] = (),
    extra_args: Iterable[str] = (),
) -> bytes:
    """Executes a Bloaty size report on some binary file(s).

    Args:
        filename: Path to the binary.
        config: Path to Bloaty config file.
        base_file: Path to a base binary. If provided, a size diff is performed.
        data_sources: List of Bloaty data sources for the report.
        extra_args: Additional command-line arguments to pass to Bloaty.

    Returns:
        Binary output of the Bloaty invocation.

    Raises:
        subprocess.CalledProcessError: The Bloaty invocation failed.
    """

    try:
        # If running from within a Bazel context, find the Bloaty executable
        # within the project's runfiles.
        # pylint: disable=import-outside-toplevel
        from python.runfiles import runfiles  # type: ignore

        r = runfiles.Create()
        bloaty_path = r.Rlocation("bloaty/bloaty")
    except ImportError:
        # Outside of Bazel, use Bloaty from the system path.
        default_bloaty = 'bloaty'
        bloaty_path = os.getenv('BLOATY_PATH', default_bloaty)

    cmd = [
        bloaty_path,
        '-c',
        config,
        '-d',
        ','.join(data_sources),
        '--domain',
        'vm',
        '-n',
        '0',
        filename,
        *extra_args,
    ]

    if base_file is not None:
        cmd.extend(['--', base_file])

    return subprocess.check_output(cmd)


def basic_size_report(
    elf: Path,
    bloaty_config: Path,
    data_sources: Iterable[str] = (),
    extra_args: Iterable[str] = (),
) -> Iterable[str]:
    """Runs a size report on an ELF file using the specified Bloaty config.

    Arguments:
        elf: The ELF binary on which to run.
        bloaty_config: Path to a Bloaty configuration file.
        data_sources: Hierarchical data sources to display.
        extra_args: Additional command line arguments forwarded to bloaty.

    Returns:
        The bloaty TSV output detailing the size report.
    """
    return (
        run_bloaty(
            str(elf.resolve()),
            str(bloaty_config.resolve()),
            data_sources=data_sources,
            extra_args=extra_args,
        )
        .decode('utf-8')
        .splitlines()
    )


class NoMemoryRegions(Exception):
    """Exception raised if an ELF does not define any memory region symbols."""

    def __init__(self, elf: Path):
        super().__init__(f'ELF {elf} does not define memory region symbols')
        self.elf = elf


def memory_regions_size_report(
    elf: Path,
    data_sources: Iterable[str] = (),
    extra_args: Iterable[str] = (),
) -> Iterable[str]:
    """Runs a size report on an ELF file using pw_bloat memory region symbols.

    Arguments:
        elf: The ELF binary on which to run.
        data_sources: Hierarchical data sources to display.
        extra_args: Additional command line arguments forwarded to bloaty.

    Returns:
        The bloaty TSV output detailing the size report.

    Raises:
        NoMemoryRegions: The ELF does not define memory region symbols.
    """
    with tempfile.NamedTemporaryFile() as bloaty_config:
        with open(elf.resolve(), "rb") as infile, open(
            bloaty_config.name, "w"
        ) as outfile:
            result = generate_bloaty_config(
                infile,
                enable_memoryregions=True,
                enable_utilization=False,
                out_file=outfile,
            )

            if not result.has_memoryregions:
                raise NoMemoryRegions(elf)

        return (
            run_bloaty(
                str(elf.resolve()),
                bloaty_config.name,
                data_sources=data_sources,
                extra_args=extra_args,
            )
            .decode('utf-8')
            .splitlines()
        )


def write_file(filename: str, contents: str, out_dir_file: str) -> None:
    path = os.path.join(out_dir_file, filename)
    with open(path, 'w') as output_file:
        output_file.write(contents)
    _LOG.debug('Output written to %s', path)


def create_binary_sizes_json(
    key_prefix: str,
    data_source_map: DataSourceMap,
    full_json: bool,
    ignore_unused_labels: bool,
) -> str:
    """Creates a binary_sizes.json file content from a list of labels.

    Args:
      key_prefix: Prefix for the json keys.
      data_source_map: Hierarchical structure containing size of sources.
      full_json: Report contains all sources, otherwise just top level.
      ignore_unused_labels: Doesn't include labels of size zero in json.

    Returns:
      A string of content to write to binary_sizes.json file.
    """
    json_content = {}
    if full_json:
        *ds_parents, last = data_source_map.get_ds_names()
        for label in data_source_map.labels():
            key = f'{key_prefix}.'
            for ds_parent, label_parent in zip(ds_parents, label.parents):
                key += f'{ds_parent}.{label_parent}.'
            key += f'{last}.{label.name}'
            if label.size != 0 or not ignore_unused_labels:
                json_content[key] = label.size
    else:
        for label in data_source_map.labels(ds_index=0):
            if label.size != 0 or not ignore_unused_labels:
                json_content[f'{key_prefix}.{label.name}'] = label.size
    return json.dumps(json_content, sort_keys=True, indent=2)


def single_target_output(
    target: str,
    bloaty_config: str,
    target_out_file: str,
    out_dir: str,
    data_sources: Iterable[str],
    extra_args: Iterable[str],
    json_key_prefix: str,
    full_json: bool,
    ignore_unused_labels: bool,
) -> int:
    """Generates size report for a single target.

    Args:
      target: The ELF binary on which to run.
      bloaty_config: Path to Bloaty config file.
      target_out_file: Output file name for the generated reports.
      out_dir: Path to write size reports to.
      data_sources: Hierarchical data sources to display.
      extra_args: Additional command-line arguments to pass to Bloaty.
      json_key_prefix: Prefix for the json keys, uses target name by default.
      full_json: Json report contains all hierarchical data source totals.

    Returns:
        Zero on success.

    Raises:
        subprocess.CalledProcessError: The Bloaty invocation failed.
    """

    try:
        single_output = run_bloaty(
            target,
            bloaty_config,
            data_sources=data_sources,
            extra_args=extra_args,
        )

    except subprocess.CalledProcessError:
        _LOG.error('%s: failed to run size report on %s', sys.argv[0], target)
        return 1

    single_tsv = single_output.decode().splitlines()
    single_report = BloatTableOutput(
        DataSourceMap.from_bloaty_tsv(single_tsv), MAX_COL_WIDTH, LineCharset
    )

    data_source_map = DataSourceMap.from_bloaty_tsv(single_tsv)
    rst_single_report = BloatTableOutput(
        data_source_map,
        MAX_COL_WIDTH,
        AsciiCharset,
        True,
    )

    single_report_table = single_report.create_table()

    # Generates contents for summary printed to binary_sizes.json
    binary_json_content = create_binary_sizes_json(
        json_key_prefix, data_source_map, full_json, ignore_unused_labels
    )

    print(single_report_table)
    write_file(target_out_file, rst_single_report.create_table(), out_dir)
    write_file(f'{target_out_file}.txt', single_report_table, out_dir)
    write_file(
        f'{target_out_file}{BINARY_SIZES_EXTENSION}',
        binary_json_content,
        out_dir,
    )

    return 0


def main() -> int:
    """Program entry point."""

    args = parse_args()
    extra_args = ['--tsv']
    data_sources = ['segment_names', 'symbols']
    gn_arg_dict = {}
    json_file = open(args.gn_arg_path)
    gn_arg_dict = json.load(json_file)
    json_key_prefix = args.json_key_prefix

    if args.single_report:
        single_binary_args = gn_arg_dict['binaries'][0]
        if single_binary_args['source_filter']:
            extra_args.extend(
                ['--source-filter', single_binary_args['source_filter']]
            )
        if single_binary_args['data_sources']:
            data_sources = single_binary_args['data_sources']

        # Use target binary name as json key prefix if none given
        if not json_key_prefix:
            json_key_prefix = single_binary_args['target']

        return single_target_output(
            single_binary_args['target'],
            single_binary_args['bloaty_config'],
            gn_arg_dict['target_name'],
            gn_arg_dict['out_dir'],
            data_sources,
            extra_args,
            json_key_prefix,
            args.full_json_summary,
            args.ignore_unused_labels,
        )

    default_data_sources = ['segment_names', 'symbols']

    diff_report = ''
    rst_diff_report = ''
    for curr_diff_binary in gn_arg_dict['binaries']:
        curr_extra_args = extra_args.copy()
        data_sources = default_data_sources

        if curr_diff_binary['source_filter']:
            curr_extra_args.extend(
                ['--source-filter', curr_diff_binary['source_filter']]
            )

        if curr_diff_binary['data_sources']:
            data_sources = curr_diff_binary['data_sources']

        try:
            single_output_base = run_bloaty(
                curr_diff_binary['base'],
                curr_diff_binary['bloaty_config'],
                data_sources=data_sources,
                extra_args=curr_extra_args,
            )

        except subprocess.CalledProcessError:
            _LOG.error(
                '%s: failed to run base size report on %s',
                sys.argv[0],
                curr_diff_binary["base"],
            )
            return 1

        try:
            single_output_target = run_bloaty(
                curr_diff_binary['target'],
                curr_diff_binary['bloaty_config'],
                data_sources=data_sources,
                extra_args=curr_extra_args,
            )

        except subprocess.CalledProcessError:
            _LOG.error(
                '%s: failed to run target size report on %s',
                sys.argv[0],
                curr_diff_binary['target'],
            )
            return 1

        if not single_output_target or not single_output_base:
            continue

        base_dsm = DataSourceMap.from_bloaty_tsv(
            single_output_base.decode().splitlines()
        )
        target_dsm = DataSourceMap.from_bloaty_tsv(
            single_output_target.decode().splitlines()
        )
        diff_dsm = target_dsm.diff(base_dsm)

        diff_report += BloatTableOutput(
            diff_dsm,
            MAX_COL_WIDTH,
            LineCharset,
            diff_label=curr_diff_binary['label'],
        ).create_table()

        curr_rst_report = RstOutput(diff_dsm, curr_diff_binary['label'])
        if rst_diff_report == '':
            rst_diff_report = curr_rst_report.create_table()
        else:
            rst_diff_report += f"{curr_rst_report.add_report_row()}\n"

    print(diff_report)
    write_file(
        gn_arg_dict['target_name'], rst_diff_report, gn_arg_dict['out_dir']
    )
    write_file(
        f"{gn_arg_dict['target_name']}.txt", diff_report, gn_arg_dict['out_dir']
    )

    return 0


if __name__ == '__main__':
    pw_cli.log.install()
    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_file/file.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12pw_file/file.proto\x12\x07pw.file\x1a\x1fpw_protobuf_protos/common.proto\"\x1b\n\x0bListRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\"\x1d\n\rDeleteRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\"\xe5\x01\n\x04Path\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x33\n\x0bpermissions\x18\x02 \x01(\x0e\x32\x19.pw.file.Path.PermissionsH\x00\x88\x01\x01\x12\x17\n\nsize_bytes\x18\x03 \x01(\rH\x01\x88\x01\x01\x12\x14\n\x07\x66ile_id\x18\x04 \x01(\rH\x02\x88\x01\x01\"@\n\x0bPermissions\x12\x08\n\x04NONE\x10\x00\x12\x08\n\x04READ\x10\x01\x12\t\n\x05WRITE\x10\x02\x12\x12\n\x0eREAD_AND_WRITE\x10\x03\x42\x0e\n\x0c_permissionsB\r\n\x0b_size_bytesB\n\n\x08_file_id\",\n\x0cListResponse\x12\x1c\n\x05paths\x18\x01 \x03(\x0b\x32\r.pw.file.Path2}\n\nFileSystem\x12\x37\n\x04List\x12\x14.pw.file.ListRequest\x1a\x15.pw.file.ListResponse\"\x00\x30\x01\x12\x36\n\x06\x44\x65lete\x12\x16.pw.file.DeleteRequest\x1a\x12.pw.protobuf.Empty\"\x00\x42\x06\x42\x04\x46ileb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_file.file_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'B\004File'
  _globals['_LISTREQUEST']._serialized_start=64
  _globals['_LISTREQUEST']._serialized_end=91
  _globals['_DELETEREQUEST']._serialized_start=93
  _globals['_DELETEREQUEST']._serialized_end=122
  _globals['_PATH']._serialized_start=125
  _globals['_PATH']._serialized_end=354
  _globals['_PATH_PERMISSIONS']._serialized_start=247
  _globals['_PATH_PERMISSIONS']._serialized_end=311
  _globals['_LISTRESPONSE']._serialized_start=356
  _globals['_LISTRESPONSE']._serialized_end=400
  _globals['_FILESYSTEM']._serialized_start=402
  _globals['_FILESYSTEM']._serialized_end=527
# @@protoc_insertion_point(module_scope)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""CLI to describe a yaml sensor definition."""

import argparse
from collections.abc import Sequence
from dataclasses import dataclass
import importlib.resources
import logging
from pathlib import Path
import subprocess
import sys

import jsonschema  # type: ignore
import jsonschema.exceptions  # type: ignore
from pw_sensor.validator import Validator
import yaml

logging.basicConfig(level=logging.DEBUG)
_LOG = logging.getLogger("sensor-describe")

_OUTPUT_SCHEMA = yaml.safe_load(
    importlib.resources.read_text("pw_sensor", "resolved_schema.json")
)


@dataclass
class Args:
    """Strongly typed wrapper around the arguments provided"""

    include_paths: Sequence[Path]
    descriptor_paths: Sequence[Path]
    generator_command: str | None
    output_file: Path | None
    log_level: int = logging.WARNING


def get_args() -> Args:
    """
    Setup the argument parser, parse the args, and return a dataclass with the
    arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--include-path",
        "-I",
        action="append",
        type=lambda p: Path(p).resolve(),
        required=True,
        help="Directories in which to search for dependency files",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="Increase verbosity level (can be used multiple times)",
    )
    parser.add_argument(
        "--generator",
        "-g",
        type=str,
        help="Generator command to run along with any flags. Data will be "
        "passed into the generator as YAML through stdin",
    )
    parser.add_argument(
        "-o",
        dest="output",
        type=Path,
        help="Write output to file instead of standard out",
    )
    parser.add_argument(
        "descriptors",
        nargs="*",
        type=lambda p: Path(p).resolve(),
        help="One or more files to validate",
    )

    args = parser.parse_args()
    if args.verbose == 0:
        log_level = logging.WARNING
    elif args.verbose == 1:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG
    return Args(
        include_paths=args.include_path,
        descriptor_paths=args.descriptors,
        generator_command=args.generator,
        output_file=args.output,
        log_level=log_level,
    )


def main() -> None:
    """
    Main entry point to the CLI. After parsing the arguments for the below
    parameters, the utility will validate the descriptor files and pass the
    output to the generator:
    - include paths
    - verbosity
    - generator
    - descriptor files
    """
    args = get_args()
    _LOG.setLevel(level=args.log_level)

    validator = Validator(
        include_paths=args.include_paths, log_level=args.log_level
    )
    superset: dict = {
        "attributes": {},
        "channels": {},
        "triggers": {},
        "units": {},
        "sensors": {},
    }
    for descriptor_file in args.descriptor_paths:
        _LOG.info("Loading '%s'", descriptor_file)
        if not descriptor_file.is_file():
            raise RuntimeError(f"'{descriptor_file}' is not a file")
        with open(descriptor_file, mode="r", encoding="utf-8") as stream:
            content = yaml.safe_load(stream=stream)
            _LOG.debug("Validating:\n%s", yaml.safe_dump(content, indent=2))
            content = validator.validate(content)
            _LOG.debug("Result:\n%s", yaml.safe_dump(content, indent=2))
        # Add sensor
        for sensor_id, values in content["sensors"].items():
            assert superset["sensors"].get(sensor_id) is None
            superset["sensors"][sensor_id] = values
        # Add channels
        for chan_id, chan_spec in content["channels"].items():
            conflict = superset["channels"].get(chan_id)
            assert conflict is None or conflict == chan_spec
            superset["channels"][chan_id] = chan_spec
        # Add attributes
        for attr_id, attr_spec in content["attributes"].items():
            conflict = superset["attributes"].get(attr_id)
            assert conflict is None or conflict == attr_spec
            superset["attributes"][attr_id] = attr_spec
        # Add triggers
        for trig_id, trig_spec in content["triggers"].items():
            conflict = superset["triggers"].get(trig_id)
            assert conflict is None or conflict == trig_spec
            superset["triggers"][trig_id] = trig_spec
        # Add units
        for units_id, units_spec in content["units"].items():
            conflict = superset["units"].get(units_id)
            assert conflict is None or conflict == units_spec
            superset["units"][units_id] = units_spec

    _LOG.debug("Final descriptor:\n%s", yaml.safe_dump(superset, indent=2))
    _LOG.info("Validating...")
    try:
        jsonschema.validate(instance=superset, schema=_OUTPUT_SCHEMA)
    except jsonschema.exceptions.ValidationError as e:
        raise RuntimeError(
            "ERROR: Malformed merged output:\n"
            f"{yaml.safe_dump(superset, indent=2)}"
        ) from e
    content_string = yaml.safe_dump(superset)

    if args.generator_command:
        cmd = args.generator_command.split(sep=" ")
        _LOG.info("Running generator %s", cmd)

        with subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as process:
            assert process.stdin is not None
            process.stdin.write(content_string.encode("utf-8"))
            out, err = process.communicate()

        if out:
            if args.output_file:
                args.output_file.parent.mkdir(parents=True, exist_ok=True)
                with open(args.output_file, mode="w", encoding="utf-8") as o:
                    o.write(out.decode("utf-8"))
            else:
                print(out.decode("utf-8"))
        if err:
            _LOG.error(err.decode("utf-8"))
        if process.returncode != 0:
            sys.exit(-1)
    else:
        print(content_string)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_compiler_protos/nested/more_nesting/test.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n:pw_protobuf_compiler_protos/nested/more_nesting/test.proto\x12\x19pw.protobuf_compiler.test\"\x18\n\x07Message\x12\r\n\x05\x66ield\x18\x01 \x01(\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_protobuf_compiler_protos.nested.more_nesting.test_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_MESSAGE']._serialized_start=89
  _globals['_MESSAGE']._serialized_end=113
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Merge multiple profraws together using llvm-profdata."""

import argparse
import enum
import json
import logging
import subprocess
import sys
from pathlib import Path
from typing import Any

_LOG = logging.getLogger(__name__)


class FailureMode(enum.Enum):
    ANY = 'any'
    ALL = 'all'

    def __str__(self):
        return self.value


def _parse_args() -> dict[str, Any]:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--llvm-profdata-path',
        type=Path,
        required=True,
        help='Path to the llvm-profdata binary to use for merging.',
    )
    parser.add_argument(
        '--test-metadata-path',
        type=Path,
        required=True,
        help='Path to the *.test_metadata.json file that describes all of the '
        'tests being used to generate a coverage report.',
    )
    parser.add_argument(
        '--profdata-path',
        type=Path,
        required=True,
        help='Path for the output merged profdata file to use with generating a'
        ' coverage report for the tests described in --test-metadata.',
    )
    parser.add_argument(
        '--depfile-path',
        type=Path,
        required=True,
        help='Path for the output depfile to convey the extra input '
        'requirements from parsing --test-metadata.',
    )
    parser.add_argument(
        '--failure-mode',
        type=FailureMode,
        choices=list(FailureMode),
        required=False,
        default=FailureMode.ANY,
        help='Sets the llvm-profdata --failure-mode option.',
    )
    return vars(parser.parse_args())


def merge_profraws(
    llvm_profdata_path: Path,
    test_metadata_path: Path,
    profdata_path: Path,
    depfile_path: Path,
    failure_mode: FailureMode,
) -> int:
    """Merge multiple profraws together using llvm-profdata."""

    # Open the test_metadata_path, parse it to JSON, and extract out the
    # profraws.
    test_metadata = json.loads(test_metadata_path.read_text())
    profraw_paths = [
        str(obj['path'])
        for obj in test_metadata
        if 'type' in obj and obj['type'] == 'profraw'
    ]

    # Generate merged profdata.
    command = [
        str(llvm_profdata_path),
        'merge',
        '--sparse',
        '--failure-mode',
        str(failure_mode),
        '-o',
        str(profdata_path),
    ] + profraw_paths

    _LOG.info('')
    _LOG.info(' '.join(command))
    _LOG.info('')

    output = subprocess.run(command)
    if output.returncode != 0:
        return output.returncode

    # Generate the depfile that describes the dependency on the profraws used to
    # create profdata_path.
    depfile_path.write_text(
        ''.join(
            [
                str(profdata_path),
                ': \\\n',
                *[str(path) + ' \\\n' for path in profraw_paths],
            ]
        )
    )

    return 0


def main() -> int:
    return merge_profraws(**_parse_args())


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generates BUILD.gn files from rules in Bazel workspace."""

import argparse
import json
import os
import re

from collections import defaultdict, deque
from pathlib import Path, PurePath, PurePosixPath
from typing import (
    Deque,
    IO,
    Iterable,
    Iterator,
    Set,
)

from pw_build.bazel_query import (
    ParseError,
    BazelLabel,
    BazelRule,
    BazelWorkspace,
)
from pw_build.gn_target import GnTarget
from pw_build.gn_writer import GnFile


class BazelToGnConverter:
    """Manages the conversion of Bazel rules into GN targets."""

    def __init__(self, pw_root: Path) -> None:
        """Instantiates a Bazel workspace.

        Args:
            pw_root: Path to Pigweed directory, e.g. "$PW_ROOT".
        """
        self._names_by_repo: dict[str, str] = {}
        self._names_by_build_arg: dict[str, str] = {}
        self._pending: Deque[BazelLabel] = deque()
        self._loaded: Set[str] = set()
        self._source_dirs: dict[str, PurePath] = {
            'pigweed': pw_root,
        }
        self._workspaces: dict[str, BazelWorkspace] = {
            'pigweed': BazelWorkspace(
                'com_google_pigweed', pw_root, fetch=False
            ),
        }
        self._revisions: dict[str, str] = {}

    def get_name(
        self,
        label: BazelLabel | None = None,
        repo: str | None = None,
        build_arg: str | None = None,
    ) -> str:
        """Returns the name of a third-party module.

        Exactly one of the "label", "repo" or "build_arg" keyword arguments must
        be provided.

        Args:
            label:      Bazel label referring to the third party module.
            repo:       Bazel repository of the third party module,
                        e.g. "com_google_foo".
            build_arg:  GN build argument of the third party module,
                        e.g. "$dir_pw_third_party_foo".
        """
        if label:
            assert not repo, 'multiple keyword arguments provided'
            repo = label.repo()
        assert not repo or not build_arg, 'multiple keyword arguments provided'
        try:
            if repo:
                return self._names_by_repo[repo]
            if build_arg:
                return self._names_by_build_arg[build_arg]
            raise AssertionError('no keyword arguments provided')
        except KeyError as e:
            raise ParseError(
                f'unrecognized third party module: "{e.args[0]}"; '
                'does it have a bazel_to_gn.json file?'
            )

    def get_source_dir(self, name: str) -> PurePath:
        """Returns the source directory for a third-party module.

        Args:
            name: Name of the third party module.
        """
        build_arg = self._build_arg(name)
        source_dir = self._source_dirs.get(build_arg)
        if not source_dir:
            raise KeyError(f'GN build argument not set: "{build_arg}"')
        return source_dir

    def parse_args_gn(self, file: IO) -> None:
        """Reads third party build arguments from args.gn.

        Args:
            file: File-like object to read from.
        """
        build_arg_pat = r'(dir_pw_third_party_\S*)\s*=\s*"([^"]*)"'
        for line in file:
            match = re.search(build_arg_pat, line)
            if match:
                build_arg = f'${match.group(1)}'
                source_dir = PurePath(match.group(2))
                self._source_dirs[build_arg] = source_dir

    def load_workspace(self, name: str, bazel_to_gn: IO) -> None:
        """Parses a bazel_to_gn.json file and loads the workspace it describes.

        Recognized fields include:
            repo:     The Bazel name of the repository.
            generate: Disables generating GN if present and set to `False`.
            targets:  A list of Bazel labels to generate GN for.
            options:  A dictionary of mapping Bazel flags to build settings.

        Args:
            name: Name of a third party module.
            bazel_to_gn: A file-like object describing the Bazel workspace.
        """
        json_data = json.load(bazel_to_gn)
        generate = json_data.get('generate', True)
        source_dir = None
        if generate:
            source_dir = self.get_source_dir(name)
        repo = json_data['repo']
        workspace = BazelWorkspace(repo, source_dir)
        workspace.generate = generate
        workspace.targets = json_data.get('targets', [])
        workspace.options = json_data.get('options', {})
        self._names_by_repo[repo] = name
        self._workspaces[name] = workspace

    def get_initial_targets(self, name: str) -> list[BazelLabel]:
        """Adds labels from a third party module to the converter queue.

        Returns the number of labels added.

        Args:
            name: Name of a previously loaded repo.
        """
        workspace = self._workspaces[name]
        repo = workspace.repo()
        self._loaded = set(workspace.targets)
        return [BazelLabel(short, repo=repo) for short in self._loaded]

    def pending(self) -> Iterable[BazelLabel]:
        """Returns the label for the next rule that needs to be loaed."""
        while self._pending:
            label = self._pending.popleft()
            if str(label) in self._loaded:
                continue
            self._loaded.add(str(label))
            yield label

    def load_rules(self, labels: list[BazelLabel]) -> Iterable[BazelRule]:
        """Queries a Bazel workspace to instantiate a rule.

        Return `None` if the GN files for the workspace are manually
        generated, otherwise returns the rule. Adds the rules deps to the queue
        of pending labels to be loaded.

        Args:
            label:  The Bazel label indicating the workspace and target.
        """
        by_repo: dict[str, list[BazelLabel]] = defaultdict(list)
        deps: Set[str] = set()
        for label in labels:
            by_repo[label.repo()].append(label)
        for repo, labels_for_repo in by_repo.items():
            name = self.get_name(repo=repo)
            workspace = self._workspaces[name]
            for rule in workspace.get_rules(labels_for_repo):
                label = rule.label()
                package = label.package()
                for attr_name in ['deps', 'implementation_deps']:
                    for dep in rule.get_list(attr_name):
                        label = BazelLabel(dep, repo=repo, package=package)
                        deps.add(str(label))
                yield rule
        self._pending.extend([BazelLabel(dep) for dep in deps])

    def package(self, rule: BazelRule) -> str:
        """Returns the relative path to the BUILD.gn corresponding to a rule.

        The relative path is relative to $dir_pw_third_party, and consists of
        the third party module name and the package portion of the Bazel label.

        Args:
            rule: The rule to get the relative path for.
        """
        label = rule.label()
        name = self.get_name(label=label)
        return f'{name}/{label.package()}'

    def convert_rule(self, rule: BazelRule) -> GnTarget:
        """Creates a GN target from a Bazel rule.

        Args:
            rule: The rule to convert into a GnTarget.
        """
        label = rule.label()
        name = self.get_name(label=label)
        if rule.kind() == 'cc_library':
            if rule.get_bool('linkstatic'):
                target_type = f'{name}_static_library'.replace('-', '_')
            else:
                target_type = f'{name}_source_set'.replace('-', '_')
        else:
            raise ParseError(f'unsupported Bazel kind: {rule.kind()}')
        gn_target = GnTarget(target_type, label.target())
        gn_target.origin = str(label)
        gn_target.attrs = {
            'public': list(self._source_relative(name, rule, 'hdrs')),
            'sources': list(self._source_relative(name, rule, 'srcs')),
            'inputs': list(
                self._source_relative(name, rule, 'additional_linker_inputs')
            ),
            'include_dirs': list(self._source_relative(name, rule, 'includes')),
            'cflags': rule.get_list('copts'),
            'public_defines': rule.get_list('defines'),
            'ldflags': rule.get_list('linkopts'),
            'defines': rule.get_list('local_defines'),
            'public_deps': list(self._build_relative(name, rule, 'deps')),
            'deps': list(
                self._build_relative(name, rule, 'implementation_deps')
            ),
        }

        return gn_target

    def num_loaded(self) -> int:
        """Returns the number of rules loaded thus far."""
        return len(self._loaded)

    def get_workspace_revisions(self) -> Iterable[str]:
        """Returns the revisions needed by each generated workspace."""
        for name, workspace in self._workspaces.items():
            if name == 'pigweed':
                continue
            if workspace.generate:
                yield f'{name:<16}: {workspace.revision()}'

    def update_pw_package(
        self, name: str, lines: Iterator[str]
    ) -> Iterable[str]:
        """Updates the third party package revision in the pw_package module.

        Args:
            lines: Contents of the existing pw_package package file.
        """
        workspace = self._workspaces[name]
        if name in self._revisions:
            revision = self._revisions[name]
        else:
            revision = workspace.revision('HEAD')
        for line in lines:
            line = line.rstrip()
            m = re.match(r'(.*commit=[\'"])([a-z0-9]*)([\'"],.*)', line)
            if not m:
                yield line
                continue
            current = m.group(2)
            if workspace.timestamp(current) < workspace.timestamp(revision):
                yield f'{m.group(1)}{revision}{m.group(3)}'
            else:
                yield line
        yield ''

    def get_imports(self, gn_target: GnTarget) -> Iterable[str]:
        """Returns the GNI files needed by the given target."""
        for build_arg in gn_target.build_args():
            name = self.get_name(build_arg=build_arg)
            yield f'$dir_pw_third_party/{name}/{name}.gni'

    def update_doc_rst(self, name: str, lines: Iterator[str]) -> Iterable[str]:
        """Replaces the "Version" part of docs.rst with the latest revision.

        This will truncate everything after the "generated section" comment and
        add the comment and version information. If the file does not have the
        comment, the comment and information will appended to the end of the
        file.

        Args:
            lines: Iterator of lines.
        """
        workspace = self._workspaces[name]
        comment = '.. DO NOT EDIT BELOW THIS LINE. Generated section.'
        url = workspace.url().rstrip('.git')
        revision = workspace.revision()
        short = revision[:8]
        for line in lines:
            line = line.rstrip()
            if line == comment:
                break
            yield line
        yield comment
        yield ''
        yield 'Version'
        yield '======='
        yield f'The update script was last run for revision `{short}`_.'
        yield ''
        yield f'.. _{short}: {url}/tree/{revision}'
        yield ''

    def _build_arg(self, name: str) -> str:
        """Returns the GN build argument for a third party module."""
        build_arg = f'$dir_pw_third_party_{name}'.replace('-', '_')
        if not build_arg in self._names_by_build_arg:
            self._names_by_build_arg[build_arg] = name
        return build_arg

    def _source_relative(
        self, name: str, rule: BazelRule, attr_name: str
    ) -> Iterable[str]:
        """Provides GN paths relative to the third party source directory."""
        if not rule.has_attr(attr_name):
            return
        attr_type = rule.attr_type(attr_name)
        build_arg = self._build_arg(name)
        repo = rule.label().repo()
        if attr_type == 'string_list':
            for item in rule.get_list(attr_name):
                yield f'{build_arg}/{item}'
        elif attr_type == 'label_list':
            for item in rule.get_list(attr_name):
                label = BazelLabel(item, repo=repo)
                yield f'{build_arg}/{label.package()}/{label.target()}'
        else:
            raise ParseError(f'unknown attribute type: {attr_type}')

    def _build_relative(
        self, name: str, rule: BazelRule, attr_name: str
    ) -> Iterable[str]:
        """Provides GN labels relative to the directory under //third_party."""
        label = rule.label()
        repo = label.repo()
        for other_str in rule.get_list(attr_name):
            other = BazelLabel(other_str, repo=repo, package=label.package())
            package = f'{name}/{label.package()}'
            other_package = f'{self.get_name(label=other)}/{other.package()}'

            # Abbreviate the label only if it is part of the same repo.
            if label.repo() != other.repo():
                path = PurePosixPath('$dir_pw_third_party', other_package)
            elif other_package == package:
                path = None
            else:
                path = PurePosixPath(package)
                other_path = PurePosixPath(other_package)
                common = PurePosixPath(
                    *os.path.commonprefix([path.parts, other_path.parts])
                )
                walk_up = PurePosixPath(
                    *(['..'] * (len(path.parts) - len(common.parts)))
                )
                walk_down = other_path.relative_to(common)
                path = PurePosixPath(walk_up, walk_down)

            if not path:
                yield f':{other.target()}'
            elif path.name == other.target():
                yield f'{path}'
            else:
                yield f'{path}:{other.target()}'

    def _get_http_archives(self) -> dict[str, BazelRule]:
        """Returns a mapping of third party modules to rules.

        The returned rules described the most recently required version of the
        third party module.
        """
        # First, examine http_archives in the third_party workspaces.
        http_archives = {}
        for name, workspace in self._workspaces.items():
            if name == 'pigweed':
                continue
            if not workspace.generate:
                continue
            for rule in workspace.get_http_archives():
                repo = rule.label().target()
                if repo not in self._names_by_repo:
                    continue
                other_name = self._names_by_repo[repo]
                other = self._workspaces[other_name]
                if not other.generate:
                    continue
                tag = rule.get_str('strip_prefix').replace(f'{other_name}-', '')
                revision = other.revision(tag)
                timestamp = other.timestamp(revision)
                if other_name in self._revisions:
                    strictest = other.timestamp(self._revisions[other_name])
                    keep = strictest < timestamp
                else:
                    keep = True
                if keep:
                    http_archives[repo] = rule
                    self._revisions[other_name] = revision

        # Next, compare them to those in the WORKSPACE file.
        pigweed = self._workspaces['pigweed']
        for rule in pigweed.get_http_archives():
            repo = rule.label().target()
            if not repo in self._names_by_repo:
                continue
            name = self._names_by_repo[repo]
            workspace = self._workspaces[name]
            if not workspace.generate:
                continue
            if name not in self._revisions:
                old_rev = rule.get_str('strip_prefix').replace(f'{name}-', '')
                new_rev = workspace.revision('HEAD')
                rule.set_attr('strip_prefix', f'{name}-{new_rev}')
                if rule.has_attr('url'):
                    url = rule.get_str('url')
                    rule.set_attr('url', url.replace(old_rev, new_rev))
                if rule.has_attr('urls'):
                    urls = rule.get_list('urls')
                    urls = [url.replace(old_rev, new_rev) for url in urls]
                    rule.set_attr('urls', urls)
                keep = True
            else:
                tag = rule.get_str('strip_prefix').replace(f'{name}-', '')
                new_rev = workspace.revision(tag)
                timestamp = workspace.timestamp(new_rev)
                strictest = workspace.timestamp(self._revisions[name])
                keep = strictest < timestamp
            if keep:
                http_archives[repo] = rule
                self._revisions[name] = new_rev

        # Next, check that the current revisions satisfy the strict revisions.
        for name, workspace in self._workspaces.items():
            if name not in self._revisions:
                continue
            needed = workspace.timestamp(self._revisions[name])
            actual = workspace.timestamp('HEAD')
            if actual < needed:
                raise RuntimeError(f'{name} must be from after {needed}.')

        # Finally, return the mapping.
        return http_archives


def _parse_args() -> argparse.Namespace:
    """Parse arguments."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '-b',
        '--build_dir',
        type=PurePath,
        help=('Build output directory, which must contain "args.gn"'),
    )
    parser.add_argument(
        'names',
        type=str,
        nargs='+',
        help=(
            'Third-party dependencies to generate GN for. '
            'Must match a subdirectoy of $PW_ROOT/third_party'
        ),
    )
    args = parser.parse_args()

    if not args.build_dir:
        pw_root = os.getenv('PW_ROOT')
        if not pw_root:
            raise RuntimeError('PW_ROOT is not set')
        args.build_dir = PurePath(pw_root, 'out')

    if not args.build_dir.is_absolute():
        args.build_dir = args.pw_root.joinpath(args.build_dir)

    return args


def _overprint(msg: str) -> None:
    """Prints with a carriage return instead of a newline."""
    print(msg.ljust(80), end='\r', flush=True)


def _bazel_to_gn(args: argparse.Namespace) -> None:
    """Generates BUILD.gn files from rules in Bazel workspace.

    This script is intended to be as unit-testable as possible. As a result,
    most functionality has been pushed into testable methods of
    BazelToGnConverter.

    This method primarily consists of three things:
      1. Print statements to provide feedback to the user.
      2. File operations, to make subroutines more unit testable.
      3. Control flow and loops around the two previous categories.

    Args:
        args: Script arguments. See `_parse_args`.
    """
    build_dir = Path(args.build_dir)
    b2g = BazelToGnConverter(build_dir.parent)

    args_gn_path = build_dir.joinpath('args.gn')
    print(f'Reading build arguments from {args_gn_path}...')
    with open(args_gn_path) as args_gn:
        b2g.parse_args_gn(args_gn)

    print('Converting Bazel rules and their dependencies to GN targets...')
    third_party_path = Path(build_dir.parent, 'third_party')
    for child in third_party_path.iterdir():
        try:
            if child.is_dir():
                with open(child.joinpath('bazel_to_gn.json')) as file:
                    b2g.load_workspace(child.name, file)
                print(f'Bazel workspace loaded for //third_party/{child.name}')
        except FileNotFoundError:
            pass

    print('Starting from:')
    for name in args.names:
        try:
            labels = b2g.get_initial_targets(name)
        except KeyError:
            print(f'E: Unable to get initial targets for "{name}".')
            print(f'E: Is "//third_party/{name}/bazel_to_gn.json" missing?')
            return
        print(f'  {len(labels)} initial rule(s) in {name}')

    by_package: dict[str, list[GnTarget]] = defaultdict(list)
    while labels:
        for rule in b2g.load_rules(labels):
            by_package[b2g.package(rule)].append(b2g.convert_rule(rule))
            _overprint(f'[{b2g.num_loaded()}] {rule.label()}')
        labels = list(b2g.pending())
    print(f'[{b2g.num_loaded()}] Conversion complete!'.ljust(80))

    for package, gn_targets in sorted(by_package.items()):
        build_gn_path = third_party_path.joinpath(package, 'BUILD.gn')
        imports = set().union(
            *[b2g.get_imports(gn_target) for gn_target in gn_targets]
        )
        _overprint(f'Writing {build_gn_path}...')
        with GnFile(build_gn_path) as build_gn:
            build_gn.write_file(imports, gn_targets)

    names = {package.split('/')[0] for package in by_package.keys()}

    for name in names:
        update_path = build_dir.parent.joinpath(
            'pw_package',
            'py',
            'pw_package',
            'packages',
            name.replace('-', '_') + '.py',
        )
        _overprint(f'Updating {update_path}...')
        with open(update_path, 'r') as pkg_file:
            contents = '\n'.join(b2g.update_pw_package(name, pkg_file))
        with open(update_path, 'w') as pkg_file:
            pkg_file.write(contents)
        print(f'Updating {update_path}  with current revision.')

    for name in names:
        update_path = third_party_path.joinpath(name, 'docs.rst')
        _overprint(f'Updating {update_path}...')
        with open(update_path, 'r') as docs_rst:
            contents = '\n'.join(b2g.update_doc_rst(name, docs_rst))
        with open(update_path, 'w') as docs_rst:
            docs_rst.write(contents)
        print(f'Updated {update_path} with current revision.')

    print('Done!')

    print(
        'Make sure to update your WORKSPACE file to fetch the following '
        + 'revisions or later:'
    )
    for revision in b2g.get_workspace_revisions():
        print(revision)


if __name__ == '__main__':
    _bazel_to_gn(_parse_args())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Common helpful decorators for Python code."""

import inspect
from pathlib import Path
from typing import Callable
import warnings


def deprecated(deprecation_note: str):
    """Deprecation decorator.

    Emits a depreciation warning when the annotated function is used.

    An additional deprecation note is required to redirect users to the
    appropriate alternative.
    """

    def _decorator_impl(func: Callable):
        def _wrapper(*args, **kwargs):
            caller = inspect.getframeinfo(inspect.currentframe().f_back)
            deprecated_func_module = inspect.getmodule(func)
            module_or_file = (
                '.'.join(
                    (
                        deprecated_func_module.__package__,
                        Path(deprecated_func_module.__file__).stem,
                    )
                )
                if deprecated_func_module.__package__ is not None
                else deprecated_func_module.__file__
            )
            warnings.warn(
                ' '.join(
                    (
                        f'{caller.filename}:{caller.lineno}:',
                        f'{module_or_file}.{func.__qualname__}()',
                        'is deprecated.',
                        deprecation_note,
                    )
                ),
                DeprecationWarning,
            )
            return func(*args, **kwargs)

        return _wrapper

    return _decorator_impl



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2023 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class I2cWriteRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BUS_INDEX_FIELD_NUMBER: builtins.int
    TARGET_ADDRESS_FIELD_NUMBER: builtins.int
    REGISTER_ADDRESS_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    bus_index: builtins.int
    """Which I2C initiator bus to communicate on."""
    target_address: builtins.int
    """7-bit I2C target address to write to."""
    register_address: builtins.bytes
    """Register address to write. Follow the endianness required by the
    responder for multi-byte address.
    """
    value: builtins.bytes
    """Value to write. Follow the endianness required by the responder."""
    def __init__(
        self,
        *,
        bus_index: builtins.int = ...,
        target_address: builtins.int = ...,
        register_address: builtins.bytes = ...,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bus_index", b"bus_index", "register_address", b"register_address", "target_address", b"target_address", "value", b"value"]) -> None: ...

global___I2cWriteRequest = I2cWriteRequest

@typing_extensions.final
class I2cWriteResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___I2cWriteResponse = I2cWriteResponse

@typing_extensions.final
class I2cReadRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BUS_INDEX_FIELD_NUMBER: builtins.int
    TARGET_ADDRESS_FIELD_NUMBER: builtins.int
    REGISTER_ADDRESS_FIELD_NUMBER: builtins.int
    READ_SIZE_FIELD_NUMBER: builtins.int
    bus_index: builtins.int
    """Which I2C initiator bus to communicate on."""
    target_address: builtins.int
    """7-bit I2C target address to read from."""
    register_address: builtins.bytes
    """Register address to write. Follow the endianness required by the
    responder for multi-byte address.
    """
    read_size: builtins.int
    """Expected number of bytes from the responder."""
    def __init__(
        self,
        *,
        bus_index: builtins.int = ...,
        target_address: builtins.int = ...,
        register_address: builtins.bytes = ...,
        read_size: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bus_index", b"bus_index", "read_size", b"read_size", "register_address", b"register_address", "target_address", b"target_address"]) -> None: ...

global___I2cReadRequest = I2cReadRequest

@typing_extensions.final
class I2cReadResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    value: builtins.bytes
    def __init__(
        self,
        *,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["value", b"value"]) -> None: ...

global___I2cReadResponse = I2cReadResponse



#!/usr/bin/env python3

# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Create transitive CLs for requirements on internal Gerrits.

This is only intended to be used by Googlers.

If the current CL needs to be tested alongside internal-project:1234 on an
internal project, but "internal-project" is something that can't be referenced
publicly, this automates creation of a CL on the pigweed-internal Gerrit that
references internal-project:1234 so the current commit effectively has a
requirement on internal-project:1234.

For more see http://go/pigweed-ci-cq-intro.
"""

import argparse
import dataclasses
import json
import logging
import os
from pathlib import Path
import re
import subprocess
import sys
import tempfile
from typing import Callable, IO, Sequence
import uuid

HELPER_GERRIT = 'pigweed-internal'
HELPER_PROJECT = 'requires-helper'
HELPER_REPO = 'sso://{}/{}'.format(HELPER_GERRIT, HELPER_PROJECT)

# Pass checks that look for "DO NOT ..." and block submission.
_DNS = ' '.join(
    (
        'DO',
        'NOT',
        'SUBMIT',
    )
)

# Subset of the output from pushing to Gerrit.
DEFAULT_OUTPUT = f'''
remote:
remote:   https://{HELPER_GERRIT}-review.git.corp.google.com/c/{HELPER_PROJECT}/+/123456789 {_DNS} [NEW]
remote:
'''.strip()

_LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class Change:
    gerrit_name: str
    number: int


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


def dump_json_patches(obj: Sequence[Change], outs: IO):
    json.dump(obj, outs, indent=2, cls=EnhancedJSONEncoder)


def log_entry_exit(func: Callable) -> Callable:
    def wrapper(*args, **kwargs):
        _LOG.debug('entering %s()', func.__name__)
        _LOG.debug('args %r', args)
        _LOG.debug('kwargs %r', kwargs)
        try:
            res = func(*args, **kwargs)
            _LOG.debug('return value %r', res)
            return res
        except Exception as exc:
            _LOG.debug('exception %r', exc)
            raise
        finally:
            _LOG.debug('exiting %s()', func.__name__)

    return wrapper


@log_entry_exit
def parse_args() -> argparse.Namespace:
    """Creates an argument parser and parses arguments."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'requirements',
        nargs='+',
        help='Requirements to be added ("<gerrit-name>:<cl-number>").',
    )
    parser.add_argument(
        '--push',
        action=argparse.BooleanOptionalAction,
        default=True,
        help=argparse.SUPPRESS,  # This option is only for debugging.
    )

    return parser.parse_args()


@log_entry_exit
def _run_command(*args, **kwargs) -> subprocess.CompletedProcess:
    kwargs.setdefault('capture_output', True)
    _LOG.debug('%s', args)
    _LOG.debug('%s', kwargs)
    res = subprocess.run(*args, **kwargs)
    _LOG.debug('%s', res.stdout)
    _LOG.debug('%s', res.stderr)
    res.check_returncode()
    return res


@log_entry_exit
def check_status() -> bool:
    res = subprocess.run(['git', 'status'], capture_output=True)
    if res.returncode:
        _LOG.error('repository not clean, commit to suppress this warning')
        return False
    return True


@log_entry_exit
def clone(requires_dir: Path) -> None:
    _LOG.info('cloning helper repository into %s', requires_dir)
    _run_command(['git', 'clone', HELPER_REPO, '.'], cwd=requires_dir)


@log_entry_exit
def create_commit(
    requires_dir: Path, requirement_strings: Sequence[str]
) -> None:
    """Create a commit in the local tree with the given requirements."""
    change_id = str(uuid.uuid4()).replace('-', '00')
    _LOG.debug('change_id %s', change_id)

    requirement_objects: list[Change] = []
    for req in requirement_strings:
        gerrit_name, number = req.split(':', 1)
        requirement_objects.append(Change(gerrit_name, int(number)))

    path = requires_dir / 'patches.json'
    _LOG.debug('path %s', path)
    with open(path, 'w') as outs:
        dump_json_patches(requirement_objects, outs)
        outs.write('\n')

    _run_command(['git', 'add', path], cwd=requires_dir)

    # TODO: b/232234662 - Don't add 'Requires:' lines to commit messages.
    commit_message = [
        f'{_DNS} {change_id[0:10]}\n\n',
        '',
        f'Change-Id: I{change_id}',
    ]
    for req in requirement_strings:
        commit_message.append(f'Requires: {req}')

    _LOG.debug('message %s', commit_message)
    _run_command(
        ['git', 'commit', '-m', '\n'.join(commit_message)],
        cwd=requires_dir,
    )

    # Not strictly necessary, only used for logging.
    _run_command(['git', 'show'], cwd=requires_dir)


@log_entry_exit
def push_commit(requires_dir: Path, push=True) -> Change:
    """Push a commit to the helper repository.

    Args:
        requires_dir: Local checkout of the helper repository.
        push: Whether to actually push or if this is a local-only test.

    Returns a Change object referencing the pushed commit.
    """

    output: str = DEFAULT_OUTPUT
    if push:
        res = _run_command(
            ['git', 'push', HELPER_REPO, '+HEAD:refs/for/main'],
            cwd=requires_dir,
        )
        output = res.stderr.decode()

    _LOG.debug('output: %s', output)
    regex = re.compile(
        f'^\\s*remote:\\s*'
        f'https://{HELPER_GERRIT}-review.(?:git.corp.google|googlesource).com/'
        f'c/{HELPER_PROJECT}/\\+/(?P<num>\\d+)\\s+',
        re.MULTILINE,
    )
    _LOG.debug('regex %r', regex)
    match = regex.search(output)
    if not match:
        raise ValueError(f"invalid output from 'git push': {output}")
    change_num = int(match.group('num'))
    _LOG.info('created %s change %s', HELPER_PROJECT, change_num)
    return Change(HELPER_GERRIT, change_num)


@log_entry_exit
def amend_existing_change(dependency: dict[str, str]) -> None:
    """Amend the current change to depend on the dependency

    Args:
        dependency: The change on which the top of the current checkout now
            depends.
    """
    git_root = Path(
        subprocess.run(
            ['git', 'rev-parse', '--show-toplevel'],
            capture_output=True,
        )
        .stdout.decode()
        .rstrip('\n')
    )
    patches_json = git_root / 'patches.json'
    _LOG.info('%s %d', patches_json, os.path.isfile(patches_json))

    patches = []
    if os.path.isfile(patches_json):
        with open(patches_json, 'r') as ins:
            patches = json.load(ins)

    patches.append(dependency)
    with open(patches_json, 'w') as outs:
        dump_json_patches(patches, outs)
        outs.write('\n')
    _LOG.info('%s %d', patches_json, os.path.isfile(patches_json))

    _run_command(['git', 'add', patches_json])
    _run_command(['git', 'commit', '--amend', '--no-edit'])


def run(requirements: Sequence[str], push: bool = True) -> int:
    """Entry point for requires."""

    if not check_status():
        return -1

    # Create directory for checking out helper repository.
    with tempfile.TemporaryDirectory() as requires_dir_str:
        requires_dir = Path(requires_dir_str)
        # Clone into helper repository.
        clone(requires_dir)
        # Make commit with requirements from command line.
        create_commit(requires_dir, requirements)
        # Push that commit and save its number.
        change = push_commit(requires_dir, push=push)
    # Add dependency on newly pushed commit on current commit.
    amend_existing_change(change)

    return 0


def main() -> int:
    return run(**vars(parse_args()))


if __name__ == '__main__':
    try:
        # If pw_cli is available, use it to initialize logs.
        from pw_cli import log

        log.install(logging.INFO)
    except ImportError:
        # If pw_cli isn't available, display log messages like a simple print.
        logging.basicConfig(format='%(message)s', level=logging.INFO)

    sys.exit(main())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_console web browser repl and log viewer startup code."""

import importlib.resources
import logging
from pathlib import Path
from typing import Iterable

import asyncio

from pw_console.console_app import PW_CONSOLE_MODULE
from pw_console.test_mode import start_fake_logger
from pw_console.web_server import pw_console_http_server


class PwConsoleWeb:
    """Class for starting console in the browser."""

    def __init__(
        self,
        global_vars=None,
        local_vars=None,
        test_mode=False,
        loggers: dict[str, Iterable[logging.Logger]] | Iterable | None = None,
        sentence_completions: dict[str, str] | None = None,
    ) -> None:
        self.test_mode = test_mode
        self.test_mode_log_loop = asyncio.new_event_loop()
        html_package_path = f'{PW_CONSOLE_MODULE}.html'
        self.html_files = {
            '/{}'.format(t): importlib.resources.read_text(html_package_path, t)
            for t in importlib.resources.contents(html_package_path)
            if Path(t).suffix in ['.css', '.html', '.js', '.json']
        }
        self.global_vars = global_vars
        self.loggers = loggers
        self.local_vars = local_vars
        self.sentence_completions = sentence_completions

    def _test_mode_log_thread_entry(self) -> None:
        """Entry point for the user code thread."""
        asyncio.set_event_loop(self.test_mode_log_loop)
        self.test_mode_log_loop.run_forever()

    def start(self) -> None:
        if self.test_mode:
            rstdoc_text = importlib.resources.read_text(
                f'{PW_CONSOLE_MODULE}.docs', 'user_guide.rst'
            )

            start_fake_logger(
                lines=rstdoc_text.splitlines(),
                log_thread_entry=self._test_mode_log_thread_entry,
                log_thread_loop=self.test_mode_log_loop,
            )
        pw_console_http_server(
            3000,
            self.html_files,
            kernel_params={
                'global_vars': self.global_vars,
                'local_vars': self.local_vars,
                'loggers': self.loggers,
                'sentence_completions': self.sentence_completions,
            },
        )



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""REPL Kernel for web console."""

import asyncio
import io
import json
import logging
import sys
import types
from typing import Any

from aiohttp.web_ws import WebSocketResponse
from prompt_toolkit.completion import (
    CompleteEvent,
    merge_completers,
    Completion,
)
from prompt_toolkit.document import Document
from ptpython.completer import PythonCompleter, Completer
from ptpython.repl import _has_coroutine_flag

from pw_console.embed import create_word_completer
from pw_console.python_logging import log_record_to_dict

_LOG = logging.getLogger(__package__)


class UnknownRequestType(Exception):
    """Exception for request with unknown or missing types."""


class UnknownRequestData(Exception):
    """Exception for request with missing data attributes."""


class MissingCallId(Exception):
    """Exception for request with missing call id."""


def process_partial_expressions(
    suggestions: list[Completion],
) -> list[Completion]:
    """
    Some completions returned are full expressions, we need to trim them.

    Example:
    if input is 'device.rp', WordCompleter suggests:
      'device.rpcs.blinky.Blinky.Blink'
    We actually need `rpcs.blinky.Blinky`
    """
    processed_completions = []
    for suggestion in suggestions:
        completion_text = suggestion.text
        start_position = suggestion.start_position

        # Handle negative start positions
        if start_position < 0:
            last_dot = completion_text.rfind('.', 0, -1 * start_position)
            if last_dot != -1:
                # We are completing a nested property
                trimmed_text = completion_text[last_dot + 1 :]
            else:
                # No dot found, return full expression
                trimmed_text = completion_text
            processed_completions.append(
                Completion(trimmed_text, 0, suggestion.display)
            )
        else:
            # Return full expression
            processed_completions.append(
                Completion(completion_text, 0, suggestion.display)
            )

    return processed_completions


def format_completions(
    all_completions: list[Completion],
) -> list[dict[str, str]]:
    all_completions = process_partial_expressions(
        # Hide private suggestions
        [
            completion
            for completion in all_completions
            if not completion.text.startswith('_')
        ]
    )

    return list(
        map(
            lambda x: {
                'text': x.text,
                'type': (
                    'keyword' if x.display[0][1].endswith('()') else 'variable'
                ),
            },
            all_completions,
        )
    )


def _format_result_output(result) -> str:
    """Return a plaintext repr of any object."""
    try:
        formatted_result = repr(result)
    except BaseException:  # pylint: disable=broad-exception-caught
        formatted_result = ''
        # Exception is handled below instead of here.
    return formatted_result


def compile_code(code: str, mode: str) -> types.CodeType:
    return compile(
        code,
        '<stdin>',
        mode,
        dont_inherit=True,
    )


class WebSocketStreamingResponder(logging.Handler):
    """Python logging handler that sends json serialized logs.

    Args:
      connection: the WebSocketResponse object to send json logs to.
      loop: The asyncio loop to run the send_str in.
    """

    def __init__(
        self,
        connection: WebSocketResponse,
        loop: asyncio.AbstractEventLoop,
        *args,
        **kwargs,
    ) -> None:
        self.connection = connection
        self.loop = loop
        self.request_ids: list[int] = []
        super().__init__(*args, **kwargs)

    def emit(self, record: logging.LogRecord) -> None:
        # Process this log record in a separate event loop.
        asyncio.run_coroutine_threadsafe(
            self._process_log(record),
            self.loop,
        )

    async def _process_log(self, record: logging.LogRecord) -> None:
        """Send the log serialized to json via the current WebSocketResponse."""
        for req_id in self.request_ids:
            await self.connection.send_str(
                json.dumps(
                    {
                        'id': req_id,
                        'streaming': True,
                        'data': {'log_line': log_record_to_dict(record)},
                    }
                )
            )


class WebKernel:
    """Web Kernel implementation."""

    def __init__(
        self,
        connection: WebSocketResponse,
        kernel_params: dict[str, Any],
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Create a new kernel for this particular websocket connection."""
        self.connection = connection
        self.kernel_params = kernel_params
        # Make sure global and local vars are not set to None.
        if kernel_params.get('global_vars', None) is None:
            self.kernel_params['global_vars'] = {}
        if kernel_params.get('local_vars', None) is None:
            self.kernel_params['local_vars'] = {}
        self.loop = loop

        self.logger_handlers: dict[str, WebSocketStreamingResponder] = {}
        self.connected = False
        python_completer = PythonCompleter(
            self.get_globals,
            self.get_locals,
            lambda: True,
        )
        all_completers: list[Completer] = [python_completer]

        if kernel_params.get('sentence_completions'):
            word_completer = create_word_completer(
                kernel_params.get('sentence_completions', {})
            )
            all_completers.append(word_completer)

        # Merge default Python completer with the new custom one.
        self.completer = merge_completers(all_completers)

    async def handle_request(self, request) -> str:
        """Handle the request from web browser."""
        try:
            parsed_request = json.loads(request)
            request_type = parsed_request.get('type', None)
            request_data = parsed_request.get('data', None)
            call_id = parsed_request.get('id', None)
            if request_type is None:
                raise UnknownRequestType(
                    'Unknown request type: {}'.format(parsed_request)
                )
            if request_data is None:
                raise UnknownRequestData(
                    'Unknown request data: {}'.format(parsed_request)
                )

            if call_id is None:
                raise MissingCallId(
                    'Missing call id: {}'.format(parsed_request)
                )

            if request_type == 'autocomplete':
                try:
                    completions = self.handle_autocompletion(
                        request_data['code'],
                        request_data['cursor_pos'],
                    )
                    return json.dumps({'id': call_id, 'data': completions})
                except KeyError as error:
                    raise KeyError(
                        (
                            'Missing data.code or data.cursor_pos attributes:'
                            '{}'.format(request_data)
                        )
                    ) from error

            if request_type == 'eval':
                try:
                    result = await self.handle_eval(request_data['code'])
                    return json.dumps({'id': call_id, 'data': result})
                except KeyError as error:
                    raise KeyError(
                        'Missing data.code attributes: {}'.format(request_data)
                    ) from error

            if request_type == 'log_source_list':
                sources = self.handle_log_source_list()
                return json.dumps({'id': call_id, 'data': sources})
            if request_type == 'log_source_subscribe':
                try:
                    # Close requests have same call id with just .close = true
                    if 'close' in parsed_request:
                        has_unsubbed = self.handle_log_source_unsubscribe(
                            request_data['name'], call_id
                        )
                        return json.dumps(
                            {
                                'id': call_id,
                                'streaming': True,
                                'data': has_unsubbed,
                            }
                        )
                    has_subscribed = self.handle_log_source_subscribe(
                        request_data['name'], call_id
                    )
                    return json.dumps(
                        {
                            'id': call_id,
                            'streaming': True,
                            'data': has_subscribed,
                        }
                    )
                except KeyError as error:
                    raise KeyError(
                        'Missing data.name attributes: {}'.format(request_data)
                    ) from error

            return 'unknown'
        except ValueError as error:
            _LOG.error('Failed to parse request: %s', request)
            return ''

    async def handle_eval(self, code: str) -> dict[str, str] | None:
        """Evaluate user code and return output."""
        # Patch stdout and stderr to capture repl print() statements.
        temp_stdout = io.StringIO()
        temp_stderr = io.StringIO()
        original_stdout = sys.stdout
        original_stderr = sys.stderr

        sys.stdout = temp_stdout
        sys.stderr = temp_stderr

        def return_result_with_stdout_stderr(result) -> dict[str, str]:
            # Always restore original stdout and stderr
            sys.stdout = original_stdout
            sys.stderr = original_stderr

            return {
                'result': (
                    _format_result_output(result) if not result is None else ''
                ),
                'stdout': temp_stdout.getvalue(),
                'stderr': temp_stderr.getvalue(),
            }

        try:
            result = await self._eval_async(code)
        except KeyboardInterrupt:
            return_result_with_stdout_stderr(None)
            raise
        except SystemExit:
            return None
        except BaseException as e:  # pylint: disable=broad-exception-caught
            return return_result_with_stdout_stderr(_format_result_output(e))
        else:
            # Print.
            return return_result_with_stdout_stderr(result)

    async def _eval_async(self, code: str) -> Any:
        """
        Evaluate the code and return result
        """

        # WORKAROUND: Due to a bug in Jedi, the current directory is removed
        # from sys.path. See: https://github.com/davidhalter/jedi/issues/1148
        if '' not in sys.path:
            sys.path.insert(0, '')

        # Try eval first
        try:
            compiled_code = compile_code(code, 'eval')
        except SyntaxError:
            pass
        else:
            # No syntax errors for eval. Do eval.
            result = eval(  # pylint: disable=eval-used
                code, self.get_globals(), self.get_locals()
            )

            if _has_coroutine_flag(compiled_code):
                result = await result

            return result

        # If not a valid `eval` expression, compile as `exec` expression
        compiled_code = compile_code(code, 'exec')
        result = eval(  # pylint: disable=eval-used
            compiled_code, self.get_globals(), self.get_locals()
        )

        if _has_coroutine_flag(compiled_code):
            result = await result
            return result

        return

    def handle_autocompletion(
        self, code: str, cursor_pos: int
    ) -> list[dict[str, str]]:
        doc = Document(code, cursor_pos)
        all_completions = list(
            self.completer.get_completions(
                doc,
                CompleteEvent(completion_requested=False, text_inserted=True),
            )
        )
        return format_completions(all_completions)

    def handle_disconnect(self) -> None:
        _LOG.info('pw_console.web_kernel disconnecting.')
        self.connected = False
        # Clean up all log handlers as we are shutting down
        for logger_name in self.kernel_params['loggers'].keys():
            for logger in self.kernel_params['loggers'][logger_name]:
                logger.removeHandler(self.logger_handlers[logger_name])

    def get_globals(self) -> dict[str, Any]:
        return self.kernel_params.get('global_vars', globals())

    def get_locals(self) -> dict[str, Any]:
        return self.kernel_params.get('local_vars', self.get_globals())

    def handle_log_source_list(self) -> list[str]:
        if 'loggers' in self.kernel_params:
            return list(self.kernel_params['loggers'].keys())
        return []

    def handle_log_source_subscribe(self, logger_name, request_id) -> bool:
        if self.kernel_params['loggers'][logger_name]:
            if not logger_name in self.logger_handlers:
                self.logger_handlers[logger_name] = WebSocketStreamingResponder(
                    self.connection,
                    self.loop,
                )
                for logger in self.kernel_params['loggers'][logger_name]:
                    logger.addHandler(self.logger_handlers[logger_name])
            self.logger_handlers[logger_name].request_ids.append(request_id)
            return True
        return False

    def handle_log_source_unsubscribe(self, logger_name, request_id) -> bool:
        if (
            self.kernel_params['loggers'][logger_name]
            and self.logger_handlers[logger_name]
        ):
            self.logger_handlers[logger_name].request_ids.remove(request_id)
            # Remove handler if all requests have unsubscribed
            if len(self.logger_handlers[logger_name].request_ids) == 0:
                for logger in self.kernel_params['loggers'][logger_name]:
                    logger.removeHandler(self.logger_handlers[logger_name])
                del self.logger_handlers[logger_name]
            return True
        return False



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Client for the pw_transfer service, which transmits data over pw_rpc."""

import asyncio
import ctypes
import logging
import threading
from typing import Any, Callable

from pw_rpc.callback_client import BidirectionalStreamingCall
from pw_status import Status

from pw_transfer.transfer import (
    ProgressCallback,
    ProtocolVersion,
    ReadTransfer,
    Transfer,
    WriteTransfer,
)
from pw_transfer.chunk import Chunk
from pw_transfer import transfer_pb2

_LOG = logging.getLogger(__package__)

_TransferDict = dict[int, Transfer]


class _TransferStream:
    def __init__(
        self,
        method,
        chunk_handler: Callable[[Chunk], Any],
        error_handler: Callable[[Status], Any],
        max_reopen_attempts=3,
    ):
        self._method = method
        self._chunk_handler = chunk_handler
        self._error_handler = error_handler
        self._call: BidirectionalStreamingCall | None = None
        self._reopen_attempts = 0
        self._max_reopen_attempts = max_reopen_attempts

    def is_open(self) -> bool:
        return self._call is not None

    def open(self, force: bool = False) -> None:
        if force or self._call is None:
            self._call = self._method.invoke(
                lambda _, chunk: self._on_chunk_received(chunk),
                on_error=lambda _, status: self._on_stream_error(status),
            )

    def close(self) -> None:
        if self._call is not None:
            self._call.cancel()
            self._call = None

    def send(self, chunk: Chunk) -> None:
        assert self._call is not None
        self._call.send(chunk.to_message())

    def _on_chunk_received(self, chunk: Chunk) -> None:
        self._reopen_attempts = 0
        self._chunk_handler(chunk)

    def _on_stream_error(self, rpc_status: Status) -> None:
        if rpc_status is Status.FAILED_PRECONDITION:
            # FAILED_PRECONDITION indicates that the stream packet was not
            # recognized as the stream is not open. Attempt to re-open the
            # stream automatically.
            self.open(force=True)
        else:
            # Other errors are unrecoverable; clear the stream.
            _LOG.error('Transfer stream shut down with status %s', rpc_status)
            self._call = None

        self._error_handler(rpc_status)


class Manager:  # pylint: disable=too-many-instance-attributes
    """A manager for transmitting data through an RPC TransferService.

    This should be initialized with an active Manager over an RPC channel. Only
    one instance of this class should exist for a configured RPC TransferService
    -- the Manager supports multiple simultaneous transfers.

    When created, a Manager starts a separate thread in which transfer
    communications and events are handled.
    """

    def __init__(
        self,
        rpc_transfer_service,
        *,
        default_response_timeout_s: float = 2.0,
        initial_response_timeout_s: float = 4.0,
        max_retries: int = 3,
        max_lifetime_retries: int = 1500,
        max_chunk_size_bytes: int = 1024,
        default_protocol_version=ProtocolVersion.VERSION_TWO,
    ):
        """Initializes a Manager on top of a TransferService.

        Args:
          rpc_transfer_service: the pw_rpc transfer service client
          default_response_timeout_s: max time to wait between receiving packets
          initial_response_timeout_s: timeout for the first packet; may be
              longer to account for transfer handler initialization
          max_retries: number of times to retry a single package after a timeout
          max_lifetime_retires: Cumulative maximum number of times to retry over
              the course of the transfer before giving up.
          max_chunk_size_bytes: In a read transfer, the maximum size of data the
              server should send within a single packet.
          default_protocol_version: Version of the pw_transfer protocol to use.
              Defaults to the latest, but can be set to legacy for projects
              which use legacy devices.
        """
        self._service: Any = rpc_transfer_service
        self._default_response_timeout_s = default_response_timeout_s
        self._initial_response_timeout_s = initial_response_timeout_s
        self.max_retries = max_retries
        self.max_lifetime_retries = max_lifetime_retries
        self._max_chunk_size_bytes = max_chunk_size_bytes
        self._default_protocol_version = default_protocol_version

        # Ongoing transfers in the service by resource ID.
        self._read_transfers: _TransferDict = {}
        self._write_transfers: _TransferDict = {}
        self._next_session_id = ctypes.c_uint32(1)

        self._loop = asyncio.new_event_loop()
        # Set the event loop for the current thread.
        asyncio.set_event_loop(self._loop)

        # Queues are used for communication between the Manager context and the
        # dedicated asyncio transfer thread.
        self._new_transfer_queue: asyncio.Queue = asyncio.Queue()
        self._read_chunk_queue: asyncio.Queue = asyncio.Queue()
        self._write_chunk_queue: asyncio.Queue = asyncio.Queue()
        self._quit_event = asyncio.Event()

        self._thread = threading.Thread(
            target=self._start_event_loop_thread, daemon=True
        )

        # RPC streams for read and write transfers. These are shareable by
        # multiple transfers of the same type.
        self._read_stream = _TransferStream(
            self._service.Read,
            lambda chunk: self._loop.call_soon_threadsafe(
                self._read_chunk_queue.put_nowait, chunk
            ),
            self._on_read_error,
        )
        self._write_stream = _TransferStream(
            self._service.Write,
            lambda chunk: self._loop.call_soon_threadsafe(
                self._write_chunk_queue.put_nowait, chunk
            ),
            self._on_write_error,
        )

        self._thread.start()

    def __del__(self):
        # Notify the thread that the transfer manager is being destroyed and
        # wait for it to exit.
        if self._thread.is_alive():
            self._loop.call_soon_threadsafe(self._quit_event.set)
            self._thread.join()

    def read(
        self,
        resource_id: int,
        progress_callback: ProgressCallback | None = None,
        protocol_version: ProtocolVersion | None = None,
        chunk_timeout_s: float | None = None,
        initial_timeout_s: float | None = None,
        initial_offset: int = 0,
    ) -> bytes:
        """Receives ("downloads") data from the server.

        Args:
          resource_id: ID of the resource from which to read.
          progress_callback: Optional callback periodically invoked throughout
              the transfer with the transfer state. Can be used to provide user-
              facing status updates such as progress bars.
          protocol_version: The desired protocol version to use for this
              transfer. Defaults to the version the manager was initialized
              (typically VERSION_TWO).
          chunk_timeout_s: Timeout for any individual chunk.
          initial_timeout_s: Timeout for the first chunk, overrides
              chunk_timeout_s.
          initial_offset: Initial offset to start reading from. Must be
              supported by the transfer handler. All transfers support starting
              from 0, the default. Returned bytes will not have any padding
              related to this initial offset. No seeking is done in the transfer
              operation on the client side.

        Raises:
          Error: the transfer failed to complete
        """

        if resource_id in self._read_transfers:
            raise ValueError(
                f'Read transfer for resource {resource_id} already exists'
            )

        if protocol_version is None:
            protocol_version = self._default_protocol_version

        if protocol_version == ProtocolVersion.LEGACY and initial_offset != 0:
            raise ValueError(
                f'Unsupported transfer with offset {initial_offset} started '
                + 'with legacy protocol'
            )

        session_id = (
            resource_id
            if protocol_version is ProtocolVersion.LEGACY
            else self.assign_session_id()
        )

        if chunk_timeout_s is None:
            chunk_timeout_s = self._default_response_timeout_s

        if initial_timeout_s is None:
            initial_timeout_s = self._initial_response_timeout_s

        transfer = ReadTransfer(
            session_id,
            resource_id,
            self._read_stream.send,
            self._end_read_transfer,
            chunk_timeout_s,
            initial_timeout_s,
            self.max_retries,
            self.max_lifetime_retries,
            protocol_version,
            max_chunk_size=self._max_chunk_size_bytes,
            progress_callback=progress_callback,
            initial_offset=initial_offset,
        )
        self._start_read_transfer(transfer)

        transfer.done.wait()

        if not transfer.status.ok():
            raise Error(transfer.resource_id, transfer.status)

        return transfer.data

    def write(
        self,
        resource_id: int,
        data: bytes | str,
        progress_callback: ProgressCallback | None = None,
        protocol_version: ProtocolVersion | None = None,
        chunk_timeout_s: Any | None = None,
        initial_timeout_s: Any | None = None,
        initial_offset: int = 0,
    ) -> None:
        """Transmits ("uploads") data to the server.

        Args:
          resource_id: ID of the resource to which to write.
          data: Data to send to the server.
          progress_callback: Optional callback periodically invoked throughout
              the transfer with the transfer state. Can be used to provide user-
              facing status updates such as progress bars.
          protocol_version: The desired protocol version to use for this
              transfer. Defaults to the version the manager was initialized
              (defaults to LATEST).
          chunk_timeout_s: Timeout for any individual chunk.
          initial_timeout_s: Timeout for the first chunk, overrides
              chunk_timeout_s.
          initial_offset: Initial offset to start writing to. Must be supported
              by the transfer handler. All transfers support starting from 0,
              the default. data arg should start with the data you want to see
              starting at this initial offset on the server. No seeking is done
              in the transfer operation on the client side.

        Raises:
          Error: the transfer failed to complete
        """

        if isinstance(data, str):
            data = data.encode()

        if resource_id in self._write_transfers:
            raise ValueError(
                f'Write transfer for resource {resource_id} already exists'
            )

        if protocol_version is None:
            protocol_version = self._default_protocol_version

        if (
            protocol_version != ProtocolVersion.VERSION_TWO
            and initial_offset != 0
        ):
            raise ValueError(
                f'Unsupported transfer with offset {initial_offset} started '
                + 'with legacy protocol'
            )

        session_id = (
            resource_id
            if protocol_version is ProtocolVersion.LEGACY
            else self.assign_session_id()
        )

        if chunk_timeout_s is None:
            chunk_timeout_s = self._default_response_timeout_s

        if initial_timeout_s is None:
            initial_timeout_s = self._initial_response_timeout_s

        transfer = WriteTransfer(
            session_id,
            resource_id,
            data,
            self._write_stream.send,
            self._end_write_transfer,
            chunk_timeout_s,
            initial_timeout_s,
            self.max_retries,
            self.max_lifetime_retries,
            protocol_version,
            progress_callback=progress_callback,
            initial_offset=initial_offset,
        )
        self._start_write_transfer(transfer)

        transfer.done.wait()

        if not transfer.status.ok():
            raise Error(transfer.resource_id, transfer.status)

    def assign_session_id(self) -> int:
        new_id = self._next_session_id.value

        self._next_session_id = ctypes.c_uint32(self._next_session_id.value + 1)
        if self._next_session_id.value == 0:
            self._next_session_id = ctypes.c_uint32(1)

        return new_id

    def _start_event_loop_thread(self):
        """Entry point for event loop thread that starts an asyncio context."""
        asyncio.set_event_loop(self._loop)

        # Recreate the async communication channels in the context of the
        # running event loop.
        self._new_transfer_queue = asyncio.Queue()
        self._read_chunk_queue = asyncio.Queue()
        self._write_chunk_queue = asyncio.Queue()
        self._quit_event = asyncio.Event()

        self._loop.create_task(self._transfer_event_loop())
        self._loop.run_forever()

    async def _transfer_event_loop(self):
        """Main async event loop."""
        exit_thread = self._loop.create_task(self._quit_event.wait())
        new_transfer = self._loop.create_task(self._new_transfer_queue.get())
        read_chunk = self._loop.create_task(self._read_chunk_queue.get())
        write_chunk = self._loop.create_task(self._write_chunk_queue.get())

        while not self._quit_event.is_set():
            # Perform a select(2)-like wait for one of several events to occur.
            done, _ = await asyncio.wait(
                (exit_thread, new_transfer, read_chunk, write_chunk),
                return_when=asyncio.FIRST_COMPLETED,
            )

            if exit_thread in done:
                break

            if new_transfer in done:
                await new_transfer.result().begin()
                new_transfer = self._loop.create_task(
                    self._new_transfer_queue.get()
                )

            if read_chunk in done:
                self._loop.create_task(
                    self._handle_chunk(
                        self._read_transfers, read_chunk.result()
                    )
                )
                read_chunk = self._loop.create_task(
                    self._read_chunk_queue.get()
                )

            if write_chunk in done:
                self._loop.create_task(
                    self._handle_chunk(
                        self._write_transfers, write_chunk.result()
                    )
                )
                write_chunk = self._loop.create_task(
                    self._write_chunk_queue.get()
                )

        self._loop.stop()

    @staticmethod
    async def _handle_chunk(
        transfers: _TransferDict, message: transfer_pb2.Chunk
    ) -> None:
        """Processes an incoming chunk from a stream.

        The chunk is dispatched to an active transfer based on its ID. If the
        transfer indicates that it is complete, the provided completion callback
        is invoked.
        """

        chunk = Chunk.from_message(message)

        # Find a transfer for the chunk in the list of active transfers.
        try:
            if chunk.protocol_version is ProtocolVersion.LEGACY:
                transfer = next(
                    t
                    for t in transfers.values()
                    if t.resource_id == chunk.session_id
                )
            else:
                transfer = next(
                    t for t in transfers.values() if t.id == chunk.id()
                )
        except StopIteration:
            _LOG.error(
                'TransferManager received chunk for unknown transfer %d',
                chunk.id(),
            )
            # TODO(frolv): What should be done here, if anything?
            return

        await transfer.handle_chunk(chunk)

    def _on_read_error(self, status: Status) -> None:
        """Callback for an RPC error in the read stream."""

        for transfer in self._read_transfers.values():
            transfer.finish(Status.INTERNAL, skip_callback=True)
        self._read_transfers.clear()

        _LOG.error('Read stream shut down: %s', status)

    def _on_write_error(self, status: Status) -> None:
        """Callback for an RPC error in the write stream."""

        for transfer in self._write_transfers.values():
            transfer.finish(Status.INTERNAL, skip_callback=True)
        self._write_transfers.clear()

        _LOG.error('Write stream shut down: %s', status)

    def _start_read_transfer(self, transfer: Transfer) -> None:
        """Begins a new read transfer, opening the stream if it isn't."""

        self._read_transfers[transfer.resource_id] = transfer
        self._read_stream.open()

        _LOG.debug('Starting new read transfer %d', transfer.id)
        delay = 1
        self._loop.call_soon_threadsafe(
            self._loop.call_later,
            delay,
            self._new_transfer_queue.put_nowait,
            transfer,
        )

    def _end_read_transfer(self, transfer: Transfer) -> None:
        """Completes a read transfer."""
        del self._read_transfers[transfer.resource_id]

        if not transfer.status.ok():
            _LOG.error(
                'Read transfer %d terminated with status %s',
                transfer.id,
                transfer.status,
            )

    def _start_write_transfer(self, transfer: Transfer) -> None:
        """Begins a new write transfer, opening the stream if it isn't."""

        self._write_transfers[transfer.resource_id] = transfer
        self._write_stream.open()

        _LOG.debug('Starting new write transfer %d', transfer.id)
        delay = 1
        self._loop.call_soon_threadsafe(
            self._loop.call_later,
            delay,
            self._new_transfer_queue.put_nowait,
            transfer,
        )

    def _end_write_transfer(self, transfer: Transfer) -> None:
        """Completes a write transfer."""
        del self._write_transfers[transfer.resource_id]

        if not transfer.status.ok():
            _LOG.error(
                'Write transfer %d terminated with status %s',
                transfer.id,
                transfer.status,
            )


class Error(Exception):
    """Exception raised when a transfer fails.

    Stores the ID of the failed transfer resource and the error that occurred.
    """

    def __init__(self, resource_id: int, status: Status):
        super().__init__(f'Transfer {resource_id} failed with status {status}')
        self.resource_id = resource_id
        self.status = status
